@@ -86,7 +86,7 @@ gen_srp_cert_server_kx (gnutls_session_t session, opaque ** data)
   gnutls_datum_t signature, ddata;
   gnutls_certificate_credentials_t cred;
   gnutls_cert *apr_cert_list;
-  gnutls_privkey *apr_pkey;
+  gnutls_privkey_t apr_pkey;
   int apr_cert_list_length;
   gnutls_sign_algorithm_t sign_algo;
 
