@@ -4,9 +4,9 @@
  *
  * Author: Nikos Mavrogiannopoulos
  *
- * This file is part of GNUTLS.
+ * This file is part of GnuTLS.
  *
- * The GNUTLS library is free software; you can redistribute it and/or
+ * The GnuTLS is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public License
  * as published by the Free Software Foundation; either version 2.1 of
  * the License, or (at your option) any later version.
@@ -53,18 +53,19 @@
 #include <gnutls_num.h>
 #include <gnutls_record.h>
 #include <gnutls_buffers.h>
+#include <gnutls_mbuffers.h>
+#include <system.h>
 
 #include <errno.h>
 
-/* We need to disable gnulib's replacement wrappers to get native
-   Windows interfaces. */
-#undef recv
-#undef send
-
 #ifndef EAGAIN
 # define EAGAIN EWOULDBLOCK
 #endif
 
+/* this is the maximum number of messages allowed to queue.
+ */
+#define MAX_QUEUE 16
+
 /**
  * gnutls_transport_set_errno:
  * @session: is a #gnutls_session_t structure.
@@ -125,7 +126,7 @@ _gnutls_record_buffer_put (content_type_t type,
 			   gnutls_session_t session, opaque * data,
 			   size_t length)
 {
-  gnutls_buffer *buf;
+  gnutls_buffer_st *buf;
 
   if (length == 0)
     return 0;
@@ -155,7 +156,7 @@ _gnutls_record_buffer_put (content_type_t type,
       return GNUTLS_E_INVALID_REQUEST;
     }
 
-  if (_gnutls_buffer_append (buf, data, length) < 0)
+  if (_gnutls_buffer_append_data (buf, data, length) < 0)
     {
       gnutls_assert ();
       return GNUTLS_E_MEMORY_ERROR;
@@ -190,9 +191,11 @@ _gnutls_record_buffer_get_size (content_type_t type, gnutls_session_t session)
  * This function checks if there are any data to receive in the gnutls
  * buffers.
  *
- * Notice that you may also use select() to check for data in a TCP
+ * Note that you could also use select() to check for data in a TCP
  * connection, instead of this function.  GnuTLS leaves some data in
- * the tcp buffer in order for select to work.
+ * the tcp buffer in order for select to work. However the select() 
+ * alternative is not recommended and will be deprecated in later
+ * GnuTLS revisions.
  *
  * Returns: the size of that data or 0.
  **/
@@ -216,14 +219,14 @@ _gnutls_record_buffer_get (content_type_t type,
   switch (type)
     {
     case GNUTLS_APPLICATION_DATA:
-      _gnutls_buffer_get_data (&session->internals.application_data_buffer,
+      _gnutls_buffer_pop_data (&session->internals.application_data_buffer,
 			       data, &length);
       _gnutls_buffers_log ("BUFFER[REC][AD]: Read %d bytes of Data(%d)\n",
 			   (int) length, (int) type);
       break;
 
     case GNUTLS_HANDSHAKE:
-      _gnutls_buffer_get_data (&session->internals.handshake_data_buffer,
+      _gnutls_buffer_pop_data (&session->internals.handshake_data_buffer,
 			       data, &length);
       _gnutls_buffers_log ("BUF[REC][HD]: Read %d bytes of Data(%d)\n",
 			   (int) length, (int) type);
@@ -231,7 +234,7 @@ _gnutls_record_buffer_get (content_type_t type,
 
     case GNUTLS_INNER_APPLICATION:
 
-      _gnutls_buffer_get_data (&session->internals.ia_data_buffer, data,
+      _gnutls_buffer_pop_data (&session->internals.ia_data_buffer, data,
 			       &length);
       _gnutls_buffers_log ("BUF[REC][IA]: Read %d bytes of Data(%d)\n",
 			   (int) length, (int) type);
@@ -246,6 +249,19 @@ _gnutls_record_buffer_get (content_type_t type,
   return length;
 }
 
+inline static void reset_errno(gnutls_session_t session)
+{  
+  session->internals.errnum = 0;
+}
+
+inline static int get_errno(gnutls_session_t session)
+{
+  if (session->internals.errnum != 0)
+    return session->internals.errnum;
+  else 
+    return session->internals.errno_func(session->internals.transport_recv_ptr);
+}
+
 
 /* This function is like read. But it does not return -1 on error.
  * It does return gnutls_errno instead.
@@ -254,12 +270,11 @@ _gnutls_record_buffer_get (content_type_t type,
  */
 static ssize_t
 _gnutls_read (gnutls_session_t session, void *iptr,
-	      size_t sizeOfPtr, int flags)
+	      size_t sizeOfPtr, gnutls_pull_func pull_func)
 {
   size_t left;
   ssize_t i = 0;
   char *ptr = iptr;
-  unsigned j, x, sum = 0;
   gnutls_transport_ptr_t fd = session->internals.transport_recv_ptr;
 
   session->internals.direction = 0;
@@ -268,43 +283,13 @@ _gnutls_read (gnutls_session_t session, void *iptr,
   while (left > 0)
     {
 
-      session->internals.errnum = 0;
+      reset_errno(session);
 
-      if (session->internals._gnutls_pull_func == NULL)
-	{
-	  i = recv (GNUTLS_POINTER_TO_INT (fd), &ptr[sizeOfPtr - left],
-		    left, flags);
-#if HAVE_WINSOCK2_H
-	  if (i < 0)
-	    {
-	      int tmperr = WSAGetLastError ();
-	      switch (tmperr)
-		{
-		case WSAEWOULDBLOCK:
-		  session->internals.errnum = EAGAIN;
-		  break;
-
-		case WSAEINTR:
-		  session->internals.errnum = EINTR;
-		  break;
-
-		default:
-		  session->internals.errnum = EIO;
-		  break;
-		}
-	      WSASetLastError (tmperr);
-	    }
-#endif
-	}
-      else
-	i = session->internals._gnutls_pull_func (fd,
-						  &ptr[sizeOfPtr -
-						       left], left);
+      i = pull_func (fd, &ptr[sizeOfPtr - left], left);
 
       if (i < 0)
 	{
-	  int err = session->internals.errnum ? session->internals.errnum
-	    : errno;
+	  int err = get_errno(session);
 
 	  _gnutls_read_log ("READ: %d returned from %p, errno=%d gerrno=%d\n",
 			    (int) i, fd, errno, session->internals.errnum);
@@ -347,36 +332,74 @@ finish:
 
   if (_gnutls_log_level >= 7)
     {
-      char line[128];
-      char tmp[16];
-
-
       _gnutls_read_log ("READ: read %d bytes from %p\n",
 			(int) (sizeOfPtr - left), fd);
 
-      for (x = 0; x < ((sizeOfPtr - left) / 16) + 1; x++)
-	{
-	  line[0] = 0;
-
-	  sprintf (tmp, "%.4x - ", x);
-	  _gnutls_str_cat (line, sizeof (line), tmp);
-
-	  for (j = 0; j < 16; j++)
-	    {
-	      if (sum < (sizeOfPtr - left))
-		{
-		  sprintf (tmp, "%.2x ", ((unsigned char *) ptr)[sum++]);
-		  _gnutls_str_cat (line, sizeof (line), tmp);
-		}
-	    }
-	  _gnutls_read_log ("%s\n", line);
-	}
     }
 
   return (sizeOfPtr - left);
 }
 
 
+
+static ssize_t
+_gnutls_writev_emu (gnutls_session_t session, const giovec_t * giovec, int giovec_cnt)
+{
+  int ret, j = 0;
+  gnutls_transport_ptr_t fd = session->internals.transport_send_ptr;
+  void* iptr;
+  size_t sizeOfPtr;
+  size_t total = 0;
+
+  for (j=0;j<giovec_cnt;j++)
+    {
+      sizeOfPtr = giovec[j].iov_len;
+      iptr = giovec[j].iov_base;
+          
+      ret = session->internals.push_func (fd, iptr, sizeOfPtr);
+
+      if (ret == -1)
+        break;
+
+      total += ret;
+    }
+
+  if (total > 0)
+    return total;
+  
+  return ret;
+}
+
+static ssize_t
+_gnutls_writev (gnutls_session_t session, const giovec_t * giovec, int giovec_cnt)
+{
+  int i;
+  gnutls_transport_ptr_t fd = session->internals.transport_send_ptr;
+
+  reset_errno(session);
+
+  if (session->internals.push_func != NULL)
+    i = _gnutls_writev_emu(session, giovec, giovec_cnt);
+  else
+    i = session->internals.vec_push_func (fd, giovec, giovec_cnt);
+
+  if (i == -1)
+    {
+      int err = get_errno(session);
+_gnutls_debug_log("errno: %d\n", err);
+      if (err == EAGAIN)
+	return GNUTLS_E_AGAIN;
+      else if (err == EINTR)
+	return GNUTLS_E_INTERRUPTED;
+      else
+	{
+	  gnutls_assert ();
+	  return GNUTLS_E_PUSH_ERROR;
+	}
+    }
+  return i;
+}
+
 #define RCVLOWAT session->internals.lowat
 
 /* This function is only used with berkeley style sockets.
@@ -402,7 +425,7 @@ _gnutls_io_clear_peeked_data (gnutls_session_t session)
   sum = 0;
   do
     {				/* we need this to finish now */
-      ret = _gnutls_read (session, peekdata, RCVLOWAT - sum, 0);
+      ret = _gnutls_read (session, peekdata, RCVLOWAT - sum, session->internals.pull_func);
       if (ret > 0)
 	sum += ret;
     }
@@ -460,7 +483,7 @@ _gnutls_io_read_buffered (gnutls_session_t session, opaque ** iptr,
   /* If an external pull function is used, then do not leave
    * any data into the kernel buffer.
    */
-  if (session->internals._gnutls_pull_func != NULL)
+  if (session->internals.pull_func != system_read)
     {
       recvlowat = 0;
     }
@@ -529,7 +552,7 @@ _gnutls_io_read_buffered (gnutls_session_t session, opaque ** iptr,
    */
   if (recvdata - recvlowat > 0)
     {
-      ret = _gnutls_read (session, &buf[buf_pos], recvdata - recvlowat, 0);
+      ret = _gnutls_read (session, &buf[buf_pos], recvdata - recvlowat, session->internals.pull_func);
 
       /* return immediately if we got an interrupt or eagain
        * error.
@@ -560,7 +583,7 @@ _gnutls_io_read_buffered (gnutls_session_t session, opaque ** iptr,
    */
   if (ret == (recvdata - recvlowat) && recvlowat > 0)
     {
-      ret2 = _gnutls_read (session, &buf[buf_pos], recvlowat, MSG_PEEK);
+      ret2 = _gnutls_read (session, &buf[buf_pos], recvlowat, system_read_peek);
 
       if (ret2 < 0 && gnutls_error_is_fatal (ret2) == 0)
 	{
@@ -619,172 +642,37 @@ _gnutls_io_read_buffered (gnutls_session_t session, opaque ** iptr,
 /* This function is like write. But it does not return -1 on error.
  * It does return gnutls_errno instead.
  *
- * In case of E_AGAIN and E_INTERRUPTED errors, you must call gnutls_write_flush(),
- * until it returns ok (0).
+ * This function takes full responsibility of freeing msg->data.
  *
- * We need to push exactly the data in n, since we cannot send less
- * data. In TLS the peer must receive the whole packet in order
- * to decrypt and verify the integrity. 
+ * In case of E_AGAIN and E_INTERRUPTED errors, you must call
+ * gnutls_write_flush(), until it returns ok (0).
+ *
+ * We need to push exactly the data in msg->size, since we cannot send
+ * less data. In TLS the peer must receive the whole packet in order
+ * to decrypt and verify the integrity.
  *
  */
 ssize_t
 _gnutls_io_write_buffered (gnutls_session_t session,
-			   const void *iptr, size_t n)
+			   mbuffer_st *bufel, unsigned int mflag)
 {
-  size_t left;
-  unsigned j, x, sum = 0;
-  ssize_t retval, i;
-  const opaque *ptr;
-  gnutls_transport_ptr_t fd = session->internals.transport_send_ptr;
+  mbuffer_head_st * const send_buffer = &session->internals.record_send_buffer;
 
-  /* to know where the procedure was interrupted.
-   */
-  session->internals.direction = 1;
+  _mbuffer_enqueue (send_buffer, bufel);
 
-  ptr = iptr;
-
-  /* In case the previous write was interrupted, check if the
-   * iptr != NULL and we have data in the buffer.
-   * If this is true then return an error.
-   */
-  if (session->internals.record_send_buffer.length > 0 && iptr != NULL)
-    {
-      gnutls_assert ();
-      return GNUTLS_E_INVALID_REQUEST;
-    }
-
-  /* If data in the buffer exist
-   */
-  if (iptr == NULL)
-    {
-      gnutls_datum_t bdata;
-      /* checking is handled above */
-      _gnutls_buffer_get_datum (&session->internals.record_send_buffer,
-				&bdata,
-				session->internals.record_send_buffer.length);
-
-      ptr = bdata.data;
-      n = bdata.size;
-
-      _gnutls_write_log
-	("WRITE: Restoring old write. (%d bytes to send)\n", (int) n);
-    }
-
-  _gnutls_write_log ("WRITE: Will write %d bytes to %p.\n", (int) n, fd);
-
-  left = n;
-  while (left > 0)
-    {
-
-      session->internals.errnum = 0;
-
-      if (session->internals._gnutls_push_func == NULL)
-	{
-	  i = send (GNUTLS_POINTER_TO_INT (fd), &ptr[n - left], left, 0);
-#if HAVE_WINSOCK2_H
-	  if (i < 0)
-	    {
-	      int tmperr = WSAGetLastError ();
-	      switch (tmperr)
-		{
-		case WSAEWOULDBLOCK:
-		  session->internals.errnum = EAGAIN;
-		  break;
-
-		case WSAEINTR:
-		  session->internals.errnum = EINTR;
-		  break;
-
-		default:
-		  session->internals.errnum = EIO;
-		  break;
-		}
-	      WSASetLastError (tmperr);
-	    }
-#endif
-	}
-      else
-	i = session->internals._gnutls_push_func (fd, &ptr[n - left], left);
-
-      if (i == -1)
-	{
-	  int err = session->internals.errnum ? session->internals.errnum
-	    : errno;
-
-	  if (err == EAGAIN || err == EINTR)
-	    {
-	      session->internals.record_send_buffer_prev_size += n - left;
-
-	      retval =
-		_gnutls_buffer_append (&session->internals.record_send_buffer,
-				       &ptr[n - left], left);
-	      if (retval < 0)
-		{
-		  gnutls_assert ();
-		  return retval;
-		}
-
-	      _gnutls_write_log
-		("WRITE: Interrupted. Stored %d bytes to buffer. Already sent %d bytes.\n",
-		 (int) left, (int) (n - left));
-
-	      if (err == EAGAIN)
-		return GNUTLS_E_AGAIN;
-	      return GNUTLS_E_INTERRUPTED;
-	    }
-	  else
-	    {
-	      gnutls_assert ();
-	      return GNUTLS_E_PUSH_ERROR;
-	    }
-	}
-      left -= i;
-
-
-      if (_gnutls_log_level >= 7)
-	{
-	  char line[128];
-	  char tmp[16];
-
-
-	  _gnutls_write_log
-	    ("WRITE: wrote %d bytes to %p. Left %d bytes. Total %d bytes.\n",
-	     (int) i, fd, (int) left, (int) n);
-	  for (x = 0; x < (unsigned) ((i) / 16) + 1; x++)
-	    {
-	      line[0] = 0;
-
-	      if (sum > n - left)
-		break;
-
-	      sprintf (tmp, "%.4x - ", x);
-	      _gnutls_str_cat (line, sizeof (line), tmp);
-
-	      for (j = 0; j < 16; j++)
-		{
-		  if (sum < n - left)
-		    {
-		      sprintf (tmp, "%.2x ",
-			       ((const unsigned char *) ptr)[sum++]);
-		      _gnutls_str_cat (line, sizeof (line), tmp);
-		    }
-		  else
-		    break;
-		}
-	      _gnutls_write_log ("%s\n", line);
-	    }
-	}
-    }
-
-  retval = n + session->internals.record_send_buffer_prev_size;
-
-  session->internals.record_send_buffer.length = 0;
-  session->internals.record_send_buffer_prev_size = 0;
-
-  return retval;
+  _gnutls_write_log
+    ("WRITE: enqueued %d bytes for %p. Total %d bytes.\n",
+     (int)bufel->msg.size, session->internals.transport_recv_ptr,
+     (int)send_buffer->byte_length);
 
+  if (mflag == MBUFFER_FLUSH)
+    return _gnutls_io_write_flush (session);
+  else
+    return bufel->msg.size;
 }
 
+typedef ssize_t (*send_func)(gnutls_session_t, const giovec_t *, int);
+
 /* This function writes the data that are left in the
  * TLS write buffer (ie. because the previous write was
  * interrupted.
@@ -792,181 +680,135 @@ _gnutls_io_write_buffered (gnutls_session_t session,
 ssize_t
 _gnutls_io_write_flush (gnutls_session_t session)
 {
-  ssize_t ret;
+  gnutls_datum_t msg;
+  mbuffer_head_st * send_buffer = &session->internals.record_send_buffer;
+  int ret;
+  ssize_t sent = 0, tosend = 0;
+  giovec_t iovec[MAX_QUEUE];
+  int i = 0;
+  mbuffer_st* cur;
 
-  if (session->internals.record_send_buffer.length == 0)
-    return 0;			/* done */
+  _gnutls_write_log ("WRITE FLUSH: %d bytes in buffer.\n",
+		     (int)send_buffer->byte_length);
 
-  ret = _gnutls_io_write_buffered (session, NULL, 0);
-  _gnutls_write_log ("WRITE FLUSH: %d [buffer: %d]\n", (int) ret,
-		     (int) session->internals.record_send_buffer.length);
+  for (cur=_mbuffer_get_first(send_buffer, &msg);
+          cur != NULL;
+          cur = _mbuffer_get_next (cur, &msg))
+        {
+          iovec[i].iov_base = msg.data;
+          iovec[i++].iov_len = msg.size;
+          tosend += msg.size;
+          
+          /* we buffer up to MAX_QUEUE messages */
+          if (i>=sizeof(iovec)/sizeof(iovec[0]))
+            {
+              gnutls_assert();
+              return GNUTLS_E_INTERNAL_ERROR;
+            }
+        }
+        
+  ret = _gnutls_writev (session, iovec, i);
+  if (ret >= 0)
+    {
+          _mbuffer_remove_bytes (send_buffer, ret);
+          _gnutls_write_log ("WRITE: wrote %d bytes, %d bytes left.\n",
+			     ret, (int)send_buffer->byte_length);
 
-  return ret;
+          sent += ret;
+    }
+  else if (ret == GNUTLS_E_INTERRUPTED || ret == GNUTLS_E_AGAIN)
+    {
+          _gnutls_write_log ("WRITE interrupted: %d bytes left.\n",
+			     (int)send_buffer->byte_length);
+	  return ret;
+    }
+  else
+    {
+          _gnutls_write_log ("WRITE error: code %d, %d bytes left.\n",
+		     ret, (int)send_buffer->byte_length);
+
+          gnutls_assert ();
+          return ret;
+    }
+
+  if (sent < tosend)
+    {
+      gnutls_assert();
+      return GNUTLS_E_AGAIN;
+    }
+
+  return sent;
 }
 
 /* This function writes the data that are left in the
  * Handshake write buffer (ie. because the previous write was
  * interrupted.
+ *
  */
 ssize_t
 _gnutls_handshake_io_write_flush (gnutls_session_t session)
 {
-  ssize_t ret;
-  ret = _gnutls_handshake_io_send_int (session, 0, 0, NULL, 0);
-  if (ret < 0)
+  mbuffer_head_st * const send_buffer = &session->internals.handshake_send_buffer;
+  gnutls_datum_t msg;
+  int ret;
+  ssize_t total = 0;
+  mbuffer_st * cur;
+
+  _gnutls_write_log ("HWRITE FLUSH: %d bytes in buffer.\n",
+		     (int)send_buffer->byte_length);
+
+  for (cur = _mbuffer_get_first (send_buffer, &msg);
+       cur != NULL;
+       cur = _mbuffer_get_first (send_buffer, &msg))
     {
-      gnutls_assert ();
-      return ret;
+      ret = _gnutls_send_int (session, GNUTLS_HANDSHAKE,
+			      session->internals.handshake_send_buffer_htype,
+			      msg.data, msg.size, 0/* do not flush */);
+
+      if (ret >= 0)
+	{
+	  _mbuffer_remove_bytes (send_buffer, ret);
+
+	  _gnutls_write_log ("HWRITE: wrote %d bytes, %d bytes left.\n",
+			     ret, (int)send_buffer->byte_length);
+
+	  total += ret;
+	}
+      else
+	{
+	  _gnutls_write_log ("HWRITE error: code %d, %d bytes left.\n",
+			     ret, (int)send_buffer->byte_length);
+
+	  gnutls_assert ();
+	  return ret;
+	}
     }
 
-  _gnutls_write_log ("HANDSHAKE_FLUSH: written[1] %d bytes\n", (int) ret);
+  return _gnutls_io_write_flush (session);
 
-  if (session->internals.handshake_send_buffer.length == 0)
-    {
-      ret = session->internals.handshake_send_buffer_prev_size;	/* done */
-      session->internals.handshake_send_buffer_prev_size = 0;
-    }
-
-  return ret;
 }
 
 
 /* This is a send function for the gnutls handshake 
  * protocol. Just makes sure that all data have been sent.
+ *
  */
-ssize_t
-_gnutls_handshake_io_send_int (gnutls_session_t session,
-			       content_type_t type,
+void
+_gnutls_handshake_io_cache_int (gnutls_session_t session,
 			       gnutls_handshake_description_t htype,
-			       const void *iptr, size_t n)
+			       mbuffer_st *bufel)
 {
-  size_t left;
-  ssize_t ret = 0;
-  const opaque *ptr;
-  ssize_t retval = 0;
+  mbuffer_head_st * const send_buffer = &session->internals.handshake_send_buffer;
 
-  ptr = iptr;
+  _mbuffer_enqueue (send_buffer, bufel);
+  session->internals.handshake_send_buffer_htype = htype;
 
-  if (session->internals.handshake_send_buffer.length > 0 && ptr == NULL
-      && n == 0)
-    {
-      gnutls_datum_t bdata;
-
-      /* resuming previously interrupted write
-       */
-      gnutls_assert ();
-
-      /* checking is handled above */
-      _gnutls_buffer_get_datum (&session->internals.handshake_send_buffer,
-				&bdata,
-				session->internals.handshake_send_buffer.
-				length);
-
-      ptr = bdata.data;
-      n = bdata.size;
-
-      type = session->internals.handshake_send_buffer_type;
-      htype = session->internals.handshake_send_buffer_htype;
-
-    }
-  else if (session->internals.handshake_send_buffer.length > 0)
-    {
-      gnutls_assert ();
-      return GNUTLS_E_INTERNAL_ERROR;
-    }
-#ifdef WRITE_DEBUG
-  else
-    {
-      size_t sum = 0, x, j;
-
-      _gnutls_write_log ("HWRITE: will write %d bytes to %d.\n", n,
-			 gnutls_transport_get_ptr (session));
-      for (x = 0; x < ((n) / 16) + 1; x++)
-	{
-	  if (sum > n)
-	    break;
-
-	  _gnutls_write_log ("%.4x - ", x);
-	  for (j = 0; j < 16; j++)
-	    {
-	      if (sum < n)
-		{
-		  _gnutls_write_log ("%.2x ", ((unsigned char *) ptr)[sum++]);
-		}
-	      else
-		break;
-	    }
-	  _gnutls_write_log ("\n");
-	}
-      _gnutls_write_log ("\n");
-    }
-#endif
-
-  if (n == 0)
-    {				/* if we have no data to send */
-      gnutls_assert ();
-      return 0;
-    }
-  else if (ptr == NULL)
-    {
-      gnutls_assert ();
-      return GNUTLS_E_INTERNAL_ERROR;
-    }
-
-
-  left = n;
-  while (left > 0)
-    {
-      ret = _gnutls_send_int (session, type, htype, &ptr[n - left], left);
-
-      if (ret <= 0)
-	{
-	  if (ret == 0)
-	    {
-	      gnutls_assert ();
-	      ret = GNUTLS_E_INTERNAL_ERROR;
-	    }
-
-	  if (left > 0
-	      && (ret == GNUTLS_E_INTERRUPTED || ret == GNUTLS_E_AGAIN))
-	    {
-	      gnutls_assert ();
-
-	      retval =
-		_gnutls_buffer_append (&session->
-				       internals.handshake_send_buffer,
-				       &ptr[n - left], left);
-	      if (retval < 0)
-		{
-		  gnutls_assert ();
-		  return retval;
-		}
-
-	      session->internals.handshake_send_buffer_prev_size += n - left;
-
-	      session->internals.handshake_send_buffer_type = type;
-	      session->internals.handshake_send_buffer_htype = htype;
-
-	    }
-	  else
-	    {
-	      session->internals.handshake_send_buffer_prev_size = 0;
-	      session->internals.handshake_send_buffer.length = 0;
-	    }
-
-	  gnutls_assert ();
-	  return ret;
-	}
-      left -= ret;
-    }
-
-  retval = n + session->internals.handshake_send_buffer_prev_size;
-
-  session->internals.handshake_send_buffer.length = 0;
-  session->internals.handshake_send_buffer_prev_size = 0;
-
-  return retval;
+  _gnutls_write_log
+    ("HWRITE: enqueued %d. Total %d bytes.\n",
+     (int)bufel->msg.size,
+     (int)send_buffer->byte_length);
 
+  return;
 }
 
 /* This is a receive function for the gnutls handshake 
@@ -1004,14 +846,14 @@ _gnutls_handshake_io_recv_int (gnutls_session_t session,
 	  gnutls_assert ();
 
 	  tmp = sizeOfPtr;
-	  _gnutls_string_get_data (&session->internals.handshake_recv_buffer,
+	  _gnutls_buffer_pop_data (&session->internals.handshake_recv_buffer,
 				   iptr, &tmp);
 	  return tmp;
 	}
       gnutls_assert ();
 
       tmp = sizeOfPtr;
-      _gnutls_string_get_data (&session->internals.handshake_recv_buffer,
+      _gnutls_buffer_pop_data (&session->internals.handshake_recv_buffer,
 			       iptr, &tmp);
       left -= tmp;
 
@@ -1030,7 +872,7 @@ _gnutls_handshake_io_recv_int (gnutls_session_t session,
 	    {
 	      gnutls_assert ();
 
-	      _gnutls_buffer_append (&session->internals.
+	      _gnutls_buffer_append_data (&session->internals.
 				     handshake_recv_buffer, iptr, dsize);
 
 	      session->internals.handshake_recv_buffer_htype = htype;
@@ -1075,7 +917,7 @@ _gnutls_handshake_buffer_put (gnutls_session_t session, opaque * data,
     }
 
   _gnutls_buffers_log ("BUF[HSK]: Inserted %d bytes of Data\n", (int) length);
-  if (_gnutls_buffer_append (&session->internals.handshake_hash_buffer,
+  if (_gnutls_buffer_append_data (&session->internals.handshake_hash_buffer,
 			     data, length) < 0)
     {
       gnutls_assert ();
@@ -1134,3 +976,85 @@ _gnutls_handshake_buffer_clear (gnutls_session_t session)
 
   return 0;
 }
+
+/**
+ * gnutls_transport_set_pull_function:
+ * @pull_func: a callback function similar to read()
+ * @session: gnutls session
+ *
+ * This is the function where you set a function for gnutls to receive
+ * data.  Normally, if you use berkeley style sockets, do not need to
+ * use this function since the default (recv(2)) will probably be ok.
+ *
+ * PULL_FUNC is of the form,
+ * ssize_t (*gnutls_pull_func)(gnutls_transport_ptr_t, void*, size_t);
+ **/
+void
+gnutls_transport_set_pull_function (gnutls_session_t session,
+				    gnutls_pull_func pull_func)
+{
+  session->internals.pull_func = pull_func;
+}
+
+/**
+ * gnutls_transport_set_push_function:
+ * @push_func: a callback function similar to write()
+ * @session: gnutls session
+ *
+ * This is the function where you set a push function for gnutls to
+ * use in order to send data.  If you are going to use berkeley style
+ * sockets, you do not need to use this function since the default
+ * (send(2)) will probably be ok.  Otherwise you should specify this
+ * function for gnutls to be able to send data.
+ *
+ * PUSH_FUNC is of the form,
+ * ssize_t (*gnutls_push_func)(gnutls_transport_ptr_t, const void*, size_t);
+ **/
+void
+gnutls_transport_set_push_function (gnutls_session_t session,
+				    gnutls_push_func push_func)
+{
+  session->internals.push_func = push_func;
+  session->internals.vec_push_func = NULL;
+}
+
+/**
+ * @session: gnutls session
+ * gnutls_transport_set_push_function2:
+ * @vec_func: a callback function similar to writev()
+ *
+ * This is the function where you set a push function for gnutls to
+ * use in order to send data.  If you are going to use berkeley style
+ * sockets, you do not need to use this function since the default
+ * (send(2)) will probably be ok.  Otherwise you should specify this
+ * function for gnutls to be able to send data.
+ *
+ * PUSH_FUNC is of the form,
+ * ssize_t (*gnutls_push_func)(gnutls_transport_ptr_t, const void*, size_t);
+ **/
+void
+gnutls_transport_set_push_function2 (gnutls_session_t session,
+				    gnutls_vec_push_func vec_func)
+{
+  session->internals.push_func = NULL;
+  session->internals.vec_push_func = vec_func;
+}
+
+/**
+ * @session: gnutls session
+ * gnutls_transport_set_errno_function:
+ * @errno_func: a callback function similar to write()
+ *
+ * This is the function where you set a function to retrieve errno
+ * after a failed push or pull operation.
+ *
+ * errno_func is of the form,
+ * int (*gnutls_errno_func)(gnutls_transport_ptr_t);
+ * and should return the errno.
+ **/
+void
+gnutls_transport_set_errno_function (gnutls_session_t session,
+				    gnutls_errno_func errno_func)
+{
+  session->internals.errno_func = errno_func;
+}
