@@ -4,9 +4,9 @@
  *
  * Author: Nikos Mavrogiannopoulos
  *
- * This file is part of GNUTLS.
+ * This file is part of GnuTLS.
  *
- * The GNUTLS library is free software; you can redistribute it and/or
+ * The GnuTLS is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public License
  * as published by the Free Software Foundation; either version 2.1 of
  * the License, or (at your option) any later version.
@@ -22,6 +22,10 @@
  * USA
  *
  */
+#ifndef GNUTLS_BUFFERS_H
+# define GNUTLS_BUFFERS_H
+
+#define MBUFFER_FLUSH 1
 
 int _gnutls_record_buffer_put (content_type_t type,
 			       gnutls_session_t session, opaque * data,
@@ -36,10 +40,8 @@ ssize_t _gnutls_io_read_buffered (gnutls_session_t, opaque ** iptr,
 void _gnutls_io_clear_read_buffer (gnutls_session_t);
 int _gnutls_io_clear_peeked_data (gnutls_session_t session);
 
-ssize_t _gnutls_io_write_buffered (gnutls_session_t, const void *iptr,
-				   size_t n);
-ssize_t _gnutls_io_write_buffered2 (gnutls_session_t, const void *iptr,
-				    size_t n, const void *iptr2, size_t n2);
+ssize_t _gnutls_io_write_buffered (gnutls_session_t session,
+				   mbuffer_st *bufel, unsigned int mflag);
 
 int _gnutls_handshake_buffer_get_size (gnutls_session_t session);
 int _gnutls_handshake_buffer_put (gnutls_session_t session, opaque * data,
@@ -50,15 +52,16 @@ int _gnutls_handshake_buffer_get_ptr (gnutls_session_t session,
 				      opaque ** data_ptr, size_t * length);
 
 #define _gnutls_handshake_io_buffer_clear( session) \
-        _gnutls_buffer_clear( &session->internals.handshake_send_buffer); \
-        _gnutls_buffer_clear( &session->internals.handshake_recv_buffer); \
-        session->internals.handshake_send_buffer_prev_size = 0
+        _mbuffer_clear( &session->internals.handshake_send_buffer); \
+        _gnutls_buffer_clear( &session->internals.handshake_recv_buffer);
 
 ssize_t _gnutls_handshake_io_recv_int (gnutls_session_t, content_type_t,
 				       gnutls_handshake_description_t, void *,
 				       size_t);
-ssize_t _gnutls_handshake_io_send_int (gnutls_session_t, content_type_t,
+void _gnutls_handshake_io_cache_int (gnutls_session_t,
 				       gnutls_handshake_description_t,
-				       const void *, size_t);
+				       mbuffer_st *bufel);
 ssize_t _gnutls_io_write_flush (gnutls_session_t session);
 ssize_t _gnutls_handshake_io_write_flush (gnutls_session_t session);
+
+#endif
