@@ -27,6 +27,7 @@
 # define GNUTLS_GLOBAL_H
 
 #include <libtasn1.h>
+#include <gnutls/gnutls.h>
 
 int gnutls_is_secure_memory (const void *mem);
 
@@ -42,5 +43,6 @@ extern ASN1_TYPE _gnutls_gnutls_asn;
 
 extern gnutls_log_func _gnutls_log_func;
 extern int _gnutls_log_level;
+extern int gnutls_crypto_init(void);
 
 #endif
