@@ -34,6 +34,7 @@
 #include "gnutls_compress.h"
 #include "gnutls_cipher.h"
 #include "gnutls_buffers.h"
+#include "gnutls_mbuffers.h"
 #include "gnutls_kx.h"
 #include "gnutls_handshake.h"
 #include "gnutls_num.h"
@@ -75,6 +76,7 @@ _gnutls_handshake_hash_add_recvd (gnutls_session_t session,
 				  opaque * dataptr, uint32_t datalen);
 
 
+
 /* Clears the handshake hash buffers and handles.
  */
 void
@@ -646,13 +648,20 @@ _gnutls_handshake_hash_pending (gnutls_session_t session)
 static int
 _gnutls_send_finished (gnutls_session_t session, int again)
 {
-  uint8_t data[MAX_VERIFY_DATA_SIZE];
+  mbuffer_st *bufel;
+  opaque *data;
   int ret;
-  int data_size = 0;
-
+  size_t vdata_size = 0;
 
   if (again == 0)
     {
+      bufel = _gnutls_handshake_alloc (MAX_VERIFY_DATA_SIZE, MAX_VERIFY_DATA_SIZE);
+      if (bufel == NULL)
+	{
+	  gnutls_assert ();
+	  return GNUTLS_E_MEMORY_ERROR;
+	}
+      data = _mbuffer_get_udata_ptr(bufel);
 
       /* This is needed in order to hash all the required
        * messages.
@@ -668,13 +677,13 @@ _gnutls_send_finished (gnutls_session_t session, int again)
 	  ret =
 	    _gnutls_ssl3_finished (session,
 				   session->security_parameters.entity, data);
-	  data_size = 36;
+          _mbuffer_set_udata_size(bufel, 36);
 	}
       else
 	{			/* TLS 1.0+ */
 	  ret = _gnutls_finished (session,
 				  session->security_parameters.entity, data);
-	  data_size = 12;
+          _mbuffer_set_udata_size(bufel, 12);
 	}
 
       if (ret < 0)
@@ -682,40 +691,26 @@ _gnutls_send_finished (gnutls_session_t session, int again)
 	  gnutls_assert ();
 	  return ret;
 	}
+	
+      vdata_size = _mbuffer_get_udata_size(bufel);
 
       if (session->internals.finished_func)
-	session->internals.finished_func (session, data, data_size);
-    }
+	session->internals.finished_func (session, data, vdata_size);
 
-  /* Save data for safe renegotiation. 
-   */
-  if (data_size > MAX_VERIFY_DATA_SIZE)
-    {
-      gnutls_assert ();
-      return GNUTLS_E_UNEXPECTED_PACKET_LENGTH;
-    }
+      ret = _gnutls_ext_sr_finished( session, data, vdata_size, 0);
+      if (ret < 0)
+        {
+	  gnutls_assert();
+	  return ret;
+	}
 
-  if (session->security_parameters.entity == GNUTLS_CLIENT)
-    {
-      session->security_parameters.extensions.client_verify_data_len =
-	data_size;
-
-      memcpy (session->security_parameters.extensions.client_verify_data,
-	      data, data_size);
+      ret = _gnutls_send_handshake (session, bufel, GNUTLS_HANDSHAKE_FINISHED);
     }
   else
     {
-      session->security_parameters.extensions.server_verify_data_len =
-	data_size;
-
-      memcpy (session->security_parameters.extensions.server_verify_data,
-	      data, data_size);
+      ret = _gnutls_send_handshake (session, NULL, GNUTLS_HANDSHAKE_FINISHED);
     }
 
-  ret =
-    _gnutls_send_handshake (session, data_size ? data : NULL, data_size,
-			    GNUTLS_HANDSHAKE_FINISHED);
-
   return ret;
 }
 
@@ -729,7 +724,6 @@ _gnutls_recv_finished (gnutls_session_t session)
   int data_size;
   int ret;
   int vrfysize;
-  tls_ext_st *ext;
 
   ret =
     _gnutls_recv_handshake (session, &vrfy, &vrfysize,
@@ -787,24 +781,11 @@ _gnutls_recv_finished (gnutls_session_t session)
     }
   gnutls_free (vrfy);
 
-  /* Save peer's verify data for safe renegotiation */
-  if (data_size > MAX_VERIFY_DATA_SIZE)
+  ret = _gnutls_ext_sr_finished( session, data, data_size, 1);
+  if (ret < 0)
     {
-      gnutls_assert ();
-      return GNUTLS_E_UNEXPECTED_PACKET_LENGTH;
-    }
-
-  ext = &session->security_parameters.extensions;
-
-  if (session->security_parameters.entity == GNUTLS_CLIENT)
-    {
-      memcpy (ext->server_verify_data, data, data_size);
-      ext->server_verify_data_len = data_size;
-    }
-  else
-    {
-      memcpy (ext->client_verify_data, data, data_size);
-      ext->client_verify_data_len = data_size;
+      gnutls_assert();
+      return ret;
     }
 
   session->internals.initial_negotiation_completed = 1;
@@ -864,24 +845,28 @@ _gnutls_server_select_suite (gnutls_session_t session, opaque * data,
 
   /* First, check for safe renegotiation SCSV.
    */
-  if (session->internals.priorities.safe_renegotiation != SR_DISABLED)
+  if (session->internals.priorities.sr != SR_DISABLED)
     {
-      int offset;
+    int offset;
 
-      for (offset = 0; offset < datalen; offset += 2)
-	{
-	  /* TLS_RENEGO_PROTECTION_REQUEST = { 0x00, 0xff } */
-	  if (data[offset] == GNUTLS_RENEGO_PROTECTION_REQUEST_MAJOR &&
-	      data[offset + 1] == GNUTLS_RENEGO_PROTECTION_REQUEST_MINOR)
-	    {
-	      _gnutls_handshake_log
-		("HSK[%p]: Received safe renegotiation CS\n", session);
-	      session->internals.safe_renegotiation_received = 1;
-	      session->internals.connection_using_safe_renegotiation = 1;
-	      break;
-	    }
-	}
-    }
+    for (offset = 0; offset < datalen; offset += 2)
+      {
+	/* TLS_RENEGO_PROTECTION_REQUEST = { 0x00, 0xff } */
+	if (data[offset] == GNUTLS_RENEGO_PROTECTION_REQUEST_MAJOR &&
+	    data[offset + 1] == GNUTLS_RENEGO_PROTECTION_REQUEST_MINOR)
+	  {
+	    _gnutls_handshake_log
+	      ("HSK[%p]: Received safe renegotiation CS\n", session);
+	    retval = _gnutls_ext_sr_recv_cs(session);
+	    if (retval < 0)
+	      {
+		gnutls_assert();
+		return retval;
+	      }
+	    break;
+	  }
+      }
+  }
 
   pk_algo = _gnutls_server_find_pk_algos_in_ciphersuites (data, datalen);
 
@@ -1057,15 +1042,21 @@ static int
 _gnutls_send_empty_handshake (gnutls_session_t session,
 			      gnutls_handshake_description_t type, int again)
 {
-  opaque data = 0;
-  opaque *ptr;
+  mbuffer_st *bufel;
 
   if (again == 0)
-    ptr = &data;
+    {
+      bufel = _gnutls_handshake_alloc (0, 0);
+      if (bufel == NULL)
+	{
+	  gnutls_assert ();
+	  return GNUTLS_E_MEMORY_ERROR;
+	}
+    }
   else
-    ptr = NULL;
+    bufel = NULL;
 
-  return _gnutls_send_handshake (session, ptr, 0, type);
+  return _gnutls_send_handshake (session, bufel, type);
 }
 
 
@@ -1127,8 +1118,7 @@ _gnutls_handshake_hash_add_sent (gnutls_session_t session,
  * (until it returns ok), with NULL parameters.
  */
 int
-_gnutls_send_handshake (gnutls_session_t session, void *i_data,
-			uint32_t i_datasize,
+_gnutls_send_handshake (gnutls_session_t session, mbuffer_st *bufel,
 			gnutls_handshake_description_t type)
 {
   int ret;
@@ -1136,7 +1126,7 @@ _gnutls_send_handshake (gnutls_session_t session, void *i_data,
   uint32_t datasize;
   int pos = 0;
 
-  if (i_data == NULL && i_datasize == 0)
+  if (bufel == NULL)
     {
       /* we are resuming a previously interrupted
        * send.
@@ -1146,28 +1136,14 @@ _gnutls_send_handshake (gnutls_session_t session, void *i_data,
 
     }
 
-  if (i_data == NULL && i_datasize > 0)
-    {
-      gnutls_assert ();
-      return GNUTLS_E_INVALID_REQUEST;
-    }
-
   /* first run */
-  datasize = i_datasize + HANDSHAKE_HEADER_SIZE;
-  data = gnutls_malloc (datasize);
-  if (data == NULL)
-    {
-      gnutls_assert ();
-      return GNUTLS_E_MEMORY_ERROR;
-    }
+  data = _mbuffer_get_uhead_ptr(bufel);
+  datasize = _mbuffer_get_udata_size(bufel) + _mbuffer_get_uhead_size(bufel);
 
   data[pos++] = (uint8_t) type;
-  _gnutls_write_uint24 (i_datasize, &data[pos]);
+  _gnutls_write_uint24 ( _mbuffer_get_udata_size(bufel), &data[pos]);
   pos += 3;
 
-  if (i_datasize > 0)
-    memcpy (&data[pos], i_data, i_datasize);
-
   _gnutls_handshake_log ("HSK[%p]: %s was sent [%ld bytes]\n",
 			 session, _gnutls_handshake2str (type),
 			 (long) datasize);
@@ -1180,17 +1156,36 @@ _gnutls_send_handshake (gnutls_session_t session, void *i_data,
 	 _gnutls_handshake_hash_add_sent (session, type, data, datasize)) < 0)
       {
 	gnutls_assert ();
-	gnutls_free (data);
+	gnutls_free (bufel);
 	return ret;
       }
 
   session->internals.last_handshake_out = type;
 
-  ret =
-    _gnutls_handshake_io_send_int (session, GNUTLS_HANDSHAKE, type,
-				   data, datasize);
+  _gnutls_handshake_io_cache_int (session, type, bufel);
+    
+  switch (type) 
+    {
+      case GNUTLS_HANDSHAKE_CERTIFICATE_PKT: /* this one is followed by ServerHelloDone
+                                              * or ClientKeyExchange always.
+                                              */
+      case GNUTLS_HANDSHAKE_SERVER_KEY_EXCHANGE: /* as above */
+      case GNUTLS_HANDSHAKE_SERVER_HELLO: /* as above */
+      case GNUTLS_HANDSHAKE_CERTIFICATE_REQUEST: /* as above */
+      case GNUTLS_HANDSHAKE_NEW_SESSION_TICKET: /* followed by ChangeCipherSpec */
 
-  gnutls_free (data);
+      /* now for client Certificate, ClientKeyExchange and
+       * CertificateVerify are always followed by ChangeCipherSpec
+       */
+      case GNUTLS_HANDSHAKE_CERTIFICATE_VERIFY:
+      case GNUTLS_HANDSHAKE_CLIENT_KEY_EXCHANGE:
+        ret = 0;
+        break;
+      default:
+        /* send cached messages */
+        ret = _gnutls_handshake_io_write_flush(session);
+        break;
+    }
 
   return ret;
 }
@@ -1669,7 +1664,7 @@ _gnutls_client_check_if_resuming (gnutls_session_t session,
 			 session_id_len);
   _gnutls_handshake_log ("HSK[%p]: SessionID: %s\n", session,
 			 _gnutls_bin2hex (session_id, session_id_len, buf,
-					  sizeof (buf)));
+					  sizeof (buf), NULL));
 
   if (session_id_len > 0 &&
       session->internals.resumed_security_parameters.session_id_size ==
@@ -1889,6 +1884,13 @@ _gnutls_copy_ciphersuites (gnutls_session_t session,
       /* Safe renegotiation signalling CS value is { 0x00, 0xff } */
       ret_data[pos++] = 0x00;
       ret_data[pos++] = 0xff;
+      ret = _gnutls_ext_sr_send_cs(session);
+      if (ret < 0) 
+        {
+          gnutls_assert();
+          gnutls_free (cipher_suites);
+          return ret;
+        }
     }
 
   gnutls_free (cipher_suites);
@@ -1941,13 +1943,14 @@ _gnutls_copy_comp_methods (gnutls_session_t session,
 /* This should be sufficient by now. It should hold all the extensions
  * plus the headers in a hello message.
  */
-#define MAX_EXT_DATA_LENGTH 1024
+#define MAX_EXT_DATA_LENGTH 32*1024
 
 /* This function sends the client hello handshake message.
  */
 static int
 _gnutls_send_client_hello (gnutls_session_t session, int again)
 {
+  mbuffer_st *bufel = NULL;
   opaque *data = NULL;
   int extdatalen;
   int pos = 0, type;
@@ -1971,22 +1974,19 @@ _gnutls_send_client_hello (gnutls_session_t session, int again)
       /* 2 for version, (4 for unix time + 28 for random bytes==GNUTLS_RANDOM_SIZE) 
        */
 
-      data = gnutls_malloc (datalen);
-      if (data == NULL)
+      bufel = _gnutls_handshake_alloc (datalen, datalen+MAX_EXT_DATA_LENGTH);
+      if (bufel == NULL)
 	{
 	  gnutls_assert ();
 	  return GNUTLS_E_MEMORY_ERROR;
 	}
+      data = _mbuffer_get_udata_ptr(bufel);
+      extdatalen = MAX_EXT_DATA_LENGTH;
 
-      extdatalen = MAX_EXT_DATA_LENGTH
-	+
-	session->internals.resumed_security_parameters.
-	extensions.session_ticket_len;
       extdata = gnutls_malloc (extdatalen);
       if (extdata == NULL)
 	{
 	  gnutls_assert ();
-	  gnutls_free (data);
 	  return GNUTLS_E_MEMORY_ERROR;
 	}
 
@@ -2009,8 +2009,8 @@ _gnutls_send_client_hello (gnutls_session_t session, int again)
       if (hver == GNUTLS_VERSION_UNKNOWN || hver == 0)
 	{
 	  gnutls_assert ();
-	  gnutls_free (data);
-	  gnutls_free (extdata);
+	  gnutls_free (bufel);
+	  gnutls_free(extdata);
 	  return GNUTLS_E_INTERNAL_ERROR;
 	}
 
@@ -2021,14 +2021,9 @@ _gnutls_send_client_hello (gnutls_session_t session, int again)
        * (RSA uses it).
        */
       _gnutls_set_adv_version (session, hver);
+      _gnutls_set_current_version (session, hver);
 
       if (session->internals.priorities.ssl3_record_version)
-	{
-	  /* Honor the SSL3_RECORD_VERSION option
-	   */
-	  _gnutls_set_current_version (session, GNUTLS_SSL3);
-	}
-      else
 	{
 	  /* Some old implementations do not interoperate if we send a
 	   * different version in the record layer.
@@ -2038,7 +2033,7 @@ _gnutls_send_client_hello (gnutls_session_t session, int again)
 	   * handshake packet and ignore the one in the packet's record 
 	   * header.
 	   */
-	  _gnutls_set_current_version (session, hver);
+	  _gnutls_record_set_default_version (session, 3, 0);
 	}
 
       /* In order to know when this session was initiated.
@@ -2086,25 +2081,20 @@ _gnutls_send_client_hello (gnutls_session_t session, int again)
 
       if (ret > 0)
 	{
-	  datalen += ret;
-	  data = gnutls_realloc_fast (data, datalen);
-	  if (data == NULL)
+	  ret = _mbuffer_append_data (bufel, extdata, ret);
+	  if (ret < 0)
 	    {
 	      gnutls_assert ();
 	      gnutls_free (extdata);
-	      return GNUTLS_E_MEMORY_ERROR;
+	      return ret;
 	    }
-
-	  memcpy (&data[pos], extdata, ret);
-	  pos += ret;
-
 	}
       else
 	{
-	  if (ret == 0)
-	    ret = GNUTLS_E_INTERNAL_ERROR;
-	  gnutls_free (data);
-	  gnutls_free (extdata);
+	  if (extdatalen == 0)
+	    extdatalen = GNUTLS_E_INTERNAL_ERROR;
+	  gnutls_free (bufel);
+	  gnutls_free(extdata);
 	  gnutls_assert ();
 	  return ret;
 	}
@@ -2115,25 +2105,20 @@ _gnutls_send_client_hello (gnutls_session_t session, int again)
       ret = _gnutls_copy_comp_methods (session, extdata, extdatalen);
       if (ret > 0)
 	{
-	  datalen += ret;
-	  data = gnutls_realloc_fast (data, datalen);
-	  if (data == NULL)
+	  ret = _mbuffer_append_data (bufel, extdata, ret);
+	  if (ret < 0)
 	    {
 	      gnutls_assert ();
 	      gnutls_free (extdata);
-	      return GNUTLS_E_MEMORY_ERROR;
+	      return ret;
 	    }
-
-	  memcpy (&data[pos], extdata, ret);
-	  pos += ret;
-
 	}
       else
 	{
-	  if (ret == 0)
-	    ret = GNUTLS_E_INTERNAL_ERROR;
-	  gnutls_free (data);
-	  gnutls_free (extdata);
+	  if (extdatalen == 0)
+	    extdatalen = GNUTLS_E_INTERNAL_ERROR;
+	  gnutls_free (bufel);
+	  gnutls_free(extdata);
 	  gnutls_assert ();
 	  return ret;
 	}
@@ -2154,32 +2139,26 @@ _gnutls_send_client_hello (gnutls_session_t session, int again)
 
       if (ret > 0)
 	{
-	  datalen += ret;
-	  data = gnutls_realloc_fast (data, datalen);
-	  if (data == NULL)
+	  ret = _mbuffer_append_data (bufel, extdata, ret);
+	  if (ret < 0)
 	    {
 	      gnutls_assert ();
 	      gnutls_free (extdata);
-	      return GNUTLS_E_MEMORY_ERROR;
+	      return ret;
 	    }
-
-	  memcpy (&data[pos], extdata, ret);
-	}
+        }
       else if (ret < 0)
-	{
-	  gnutls_assert ();
-	  gnutls_free (data);
-	  gnutls_free (extdata);
-	  return ret;
-	}
-
+        {
+          gnutls_assert ();
+          gnutls_free (bufel);
+	  gnutls_free(extdata);
+          return ret;
+        }
     }
 
-  ret =
-    _gnutls_send_handshake (session, data, datalen,
-			    GNUTLS_HANDSHAKE_CLIENT_HELLO);
-  gnutls_free (extdata);
-  gnutls_free (data);
+  gnutls_free(extdata);
+
+  ret = _gnutls_send_handshake (session, bufel, GNUTLS_HANDSHAKE_CLIENT_HELLO);
 
   return ret;
 }
@@ -2187,8 +2166,9 @@ _gnutls_send_client_hello (gnutls_session_t session, int again)
 static int
 _gnutls_send_server_hello (gnutls_session_t session, int again)
 {
+  mbuffer_st *bufel = NULL;
   opaque *data = NULL;
-  opaque extdata[MAX_EXT_DATA_LENGTH];
+  opaque *extdata = NULL;
   int extdatalen;
   int pos = 0;
   int datalen, ret = 0;
@@ -2198,47 +2178,36 @@ _gnutls_send_server_hello (gnutls_session_t session, int again)
 
   datalen = 0;
 
-#ifdef ENABLE_SRP
-  if (IS_SRP_KX
-      (_gnutls_cipher_suite_get_kx_algo
-       (&session->security_parameters.current_cipher_suite)))
-    {
-      /* While resuming we cannot check the username extension since it is
-       * not available at this point. It will be copied on connection
-       * state activation.
-       */
-      if (session->internals.resumed == RESUME_FALSE &&
-	  session->security_parameters.extensions.srp_username[0] == 0)
-	{
-	  /* The peer didn't send a valid SRP extension with the
-	   * SRP username. The draft requires that we send a fatal
-	   * alert and abort.
-	   */
-	  gnutls_assert ();
-	  return GNUTLS_E_UNKNOWN_SRP_USERNAME;
-	}
-    }
-#endif
-
   if (again == 0)
     {
+    
+      extdata = gnutls_malloc(MAX_EXT_DATA_LENGTH);
+      if (extdata == NULL)
+        {
+          gnutls_assert();
+          return GNUTLS_E_MEMORY_ERROR;
+        }
+
       datalen = 2 + session_id_len + 1 + GNUTLS_RANDOM_SIZE + 3;
-      extdatalen =
-	_gnutls_gen_extensions (session, extdata, sizeof (extdata),
+      ret =
+	_gnutls_gen_extensions (session, extdata, MAX_EXT_DATA_LENGTH,
 				GNUTLS_EXT_ANY);
 
-      if (extdatalen < 0)
+      if (ret < 0)
 	{
 	  gnutls_assert ();
-	  return extdatalen;
+	  goto fail;
 	}
+      extdatalen = ret;
 
-      data = gnutls_malloc (datalen + extdatalen);
-      if (data == NULL)
+      bufel = _gnutls_handshake_alloc (datalen + extdatalen, datalen + extdatalen);
+      if (bufel == NULL)
 	{
 	  gnutls_assert ();
-	  return GNUTLS_E_MEMORY_ERROR;
+	  ret = GNUTLS_E_MEMORY_ERROR;
+	  goto fail;
 	}
+      data = _mbuffer_get_udata_ptr(bufel);
 
       data[pos++] =
 	_gnutls_version_get_major (session->security_parameters.version);
@@ -2260,7 +2229,7 @@ _gnutls_send_server_hello (gnutls_session_t session, int again)
       _gnutls_handshake_log ("HSK[%p]: SessionID: %s\n", session,
 			     _gnutls_bin2hex (session->security_parameters.
 					      session_id, session_id_len, buf,
-					      sizeof (buf)));
+					      sizeof (buf), NULL));
 
       memcpy (&data[pos],
 	      session->security_parameters.current_cipher_suite.suite, 2);
@@ -2280,11 +2249,10 @@ _gnutls_send_server_hello (gnutls_session_t session, int again)
 	}
     }
 
-  ret =
-    _gnutls_send_handshake (session, data, datalen,
-			    GNUTLS_HANDSHAKE_SERVER_HELLO);
-  gnutls_free (data);
+  ret = _gnutls_send_handshake (session, bufel, GNUTLS_HANDSHAKE_SERVER_HELLO);
 
+fail:
+  gnutls_free(extdata);
   return ret;
 }
 
@@ -2293,8 +2261,6 @@ _gnutls_send_hello (gnutls_session_t session, int again)
 {
   int ret;
 
-  session->internals.safe_renegotiation_received = 0;
-
   if (session->security_parameters.entity == GNUTLS_CLIENT)
     {
       ret = _gnutls_send_client_hello (session, again);
@@ -2316,7 +2282,6 @@ int
 _gnutls_recv_hello (gnutls_session_t session, opaque * data, int datalen)
 {
   int ret;
-  tls_ext_st *ext;
 
   if (session->security_parameters.entity == GNUTLS_CLIENT)
     {
@@ -2338,98 +2303,11 @@ _gnutls_recv_hello (gnutls_session_t session, opaque * data, int datalen)
 	}
     }
 
-  if (session->internals.priorities.safe_renegotiation == SR_DISABLED)
+  ret = _gnutls_ext_sr_verify(session);
+  if (ret < 0)
     {
-      gnutls_assert ();
-      return 0;
-    }
-
-  /* Safe renegotiation */
-  ext = &session->security_parameters.extensions;
-
-  if (session->internals.safe_renegotiation_received)
-    {
-      if ((ext->ri_extension_data_len < ext->client_verify_data_len) ||
-	  (memcmp (ext->ri_extension_data,
-		   ext->client_verify_data, ext->client_verify_data_len)))
-	{
-	  gnutls_assert ();
-	  _gnutls_handshake_log ("HSK[%p]: Safe renegotiation failed [1]\n",
-				 session);
-	  return GNUTLS_E_SAFE_RENEGOTIATION_FAILED;
-	}
-
-      if (session->security_parameters.entity == GNUTLS_CLIENT)
-	{
-	  if ((ext->ri_extension_data_len !=
-	       ext->client_verify_data_len + ext->server_verify_data_len) ||
-	      memcmp (ext->ri_extension_data + ext->client_verify_data_len,
-		      ext->server_verify_data,
-		      ext->server_verify_data_len) != 0)
-	    {
-	      gnutls_assert ();
-	      _gnutls_handshake_log
-		("HSK[%p]: Safe renegotiation failed [2]\n", session);
-	      return GNUTLS_E_SAFE_RENEGOTIATION_FAILED;
-	    }
-	}
-      else			/* Make sure there are 0 extra bytes */
-	{
-	  if (ext->ri_extension_data_len != ext->client_verify_data_len)
-	    {
-	      gnutls_assert ();
-	      _gnutls_handshake_log
-		("HSK[%p]: Safe renegotiation failed [3]\n", session);
-	      return GNUTLS_E_SAFE_RENEGOTIATION_FAILED;
-	    }
-	}
-
-      _gnutls_handshake_log ("HSK[%p]: Safe renegotiation succeeded\n",
-			     session);
-    }
-  else				/* safe renegotiation not received... */
-    {
-      if (session->internals.connection_using_safe_renegotiation)
-	{
-	  gnutls_assert ();
-	  _gnutls_handshake_log
-	    ("HSK[%p]: Peer previously asked for safe renegotiation\n",
-	     session);
-	  return GNUTLS_E_SAFE_RENEGOTIATION_FAILED;
-	}
-
-      /* Clients can't tell if it's an initial negotiation */
-      if (session->internals.initial_negotiation_completed)
-	{
-
-	  if (session->internals.priorities.safe_renegotiation < SR_PARTIAL)
-	    {
-	      _gnutls_handshake_log
-		("HSK[%p]: Allowing unsafe (re)negotiation\n", session);
-	    }
-	  else
-	    {
-	      gnutls_assert ();
-	      _gnutls_handshake_log
-		("HSK[%p]: Denying unsafe (re)negotiation\n", session);
-	      return GNUTLS_E_UNSAFE_RENEGOTIATION_DENIED;
-	    }
-	}
-      else
-	{
-	  if (session->internals.priorities.safe_renegotiation < SR_SAFE)
-	    {
-	      _gnutls_handshake_log
-		("HSK[%p]: Allowing unsafe initial negotiation\n", session);
-	    }
-	  else
-	    {
-	      gnutls_assert ();
-	      _gnutls_handshake_log
-		("HSK[%p]: Denying unsafe initial negotiation\n", session);
-	      return GNUTLS_E_SAFE_RENEGOTIATION_FAILED;
-	    }
-	}
+	gnutls_assert();
+	return ret;
     }
 
   return 0;
@@ -2452,12 +2330,25 @@ _gnutls_recv_hello (gnutls_session_t session, opaque * data, int datalen)
  *     CertificateVerify*
  *     [ChangeCipherSpec]
  *     Finished                     -------->
+ *                                                NewSessionTicket
  *                                              [ChangeCipherSpec]
  *                                  <--------             Finished
  *
  * (*): means optional packet.
  */
 
+/* Handshake when resumming session:
+ *      Client                                                Server
+ *
+ *      ClientHello                   -------->
+ *                                                      ServerHello
+ *                                               [ChangeCipherSpec]
+ *                                   <--------             Finished
+ *     [ChangeCipherSpec]
+ *     Finished                      -------->
+ * 
+ */
+
 /**
  * gnutls_rehandshake:
  * @session: is a #gnutls_session_t structure.
@@ -2605,16 +2496,16 @@ _gnutls_handshake_hash_init (gnutls_session_t session)
 static int
 _gnutls_send_supplemental (gnutls_session_t session, int again)
 {
+  mbuffer_st *bufel;
   int ret = 0;
 
   _gnutls_debug_log ("EXT[%p]: Sending supplemental data\n", session);
 
   if (again)
-    ret = _gnutls_send_handshake (session, NULL, 0,
-				  GNUTLS_HANDSHAKE_SUPPLEMENTAL);
+    ret = _gnutls_send_handshake (session, NULL, GNUTLS_HANDSHAKE_SUPPLEMENTAL);
   else
     {
-      gnutls_buffer buf;
+      gnutls_buffer_st buf;
       _gnutls_buffer_init (&buf);
 
       ret = _gnutls_gen_supplemental (session, &buf);
@@ -2624,9 +2515,18 @@ _gnutls_send_supplemental (gnutls_session_t session, int again)
 	  return ret;
 	}
 
-      ret = _gnutls_send_handshake (session, buf.data, buf.length,
-				    GNUTLS_HANDSHAKE_SUPPLEMENTAL);
+      bufel = _gnutls_handshake_alloc(buf.length, buf.length);
+      if (bufel == NULL)
+	{
+	  gnutls_assert ();
+	  return GNUTLS_E_MEMORY_ERROR;
+	}
+
+      _mbuffer_set_udata(bufel, buf.data, buf.length);
       _gnutls_buffer_clear (&buf);
+
+      ret = _gnutls_send_handshake (session, bufel,
+				    GNUTLS_HANDSHAKE_SUPPLEMENTAL);
     }
 
   return ret;
@@ -2732,14 +2632,9 @@ gnutls_handshake (gnutls_session_t session)
   return 0;
 }
 
-
 #define IMED_RET( str, ret, check_fatal) do { \
 	if (ret < 0) { \
-		/* EAGAIN and INTERRUPTED are always non-fatal */ \
-		if (ret == GNUTLS_E_AGAIN || ret == GNUTLS_E_INTERRUPTED) \
-			return ret; \
-                /* a warning alert might interrupt handshake */ \
-		if (check_fatal != 0 && ret==GNUTLS_E_WARNING_ALERT_RECEIVED) return ret; \
+		if (check_fatal != 0 && gnutls_error_is_fatal(ret)==0) return ret; \
 		gnutls_assert(); \
 		ERR( str, ret); \
 		_gnutls_handshake_hash_buffers_clear(session); \
@@ -2766,7 +2661,7 @@ _gnutls_handshake_client (gnutls_session_t session)
 					    internals.resumed_security_parameters.session_id,
 					    session->
 					    internals.resumed_security_parameters.session_id_size,
-					    buf, sizeof (buf)));
+					    buf, sizeof (buf), NULL));
 #endif
 
   switch (STATE)
@@ -2787,7 +2682,7 @@ _gnutls_handshake_client (gnutls_session_t session)
       IMED_RET ("recv hello", ret, 1);
 
     case STATE70:
-      if (session->security_parameters.extensions.do_recv_supplemental)
+      if (session->security_parameters.do_recv_supplemental)
 	{
 	  ret = _gnutls_recv_supplemental (session);
 	  STATE = STATE70;
@@ -2828,7 +2723,7 @@ _gnutls_handshake_client (gnutls_session_t session)
       IMED_RET ("recv server hello done", ret, 1);
 
     case STATE71:
-      if (session->security_parameters.extensions.do_send_supplemental)
+      if (session->security_parameters.do_send_supplemental)
 	{
 	  ret = _gnutls_send_supplemental (session, AGAIN (STATE71));
 	  STATE = STATE71;
@@ -2872,6 +2767,7 @@ static int
 _gnutls_send_handshake_final (gnutls_session_t session, int init)
 {
   int ret = 0;
+  int oldstate = STATE;
 
   /* Send the CHANGE CIPHER SPEC PACKET */
 
@@ -2879,15 +2775,24 @@ _gnutls_send_handshake_final (gnutls_session_t session, int init)
     {
     case STATE0:
     case STATE20:
-      ret = _gnutls_send_change_cipher_spec (session, AGAIN (STATE20));
+
       STATE = STATE20;
+
+      ret = _gnutls_handshake_io_write_flush(session);
+      if (ret < 0) 
+        {
+          gnutls_assert();
+          return ret;
+        }
+
+      ret = _gnutls_send_change_cipher_spec (session, AGAIN2 (oldstate, STATE20));
+
       if (ret < 0)
 	{
 	  ERR ("send ChangeCipherSpec", ret);
 	  gnutls_assert ();
 	  return ret;
 	}
-
       /* Initialize the connection session (start encryption) - in case of client 
        */
       if (init == TRUE)
@@ -3011,7 +2916,7 @@ _gnutls_handshake_server (gnutls_session_t session)
       IMED_RET ("send hello", ret, 0);
 
     case STATE70:
-      if (session->security_parameters.extensions.do_send_supplemental)
+      if (session->security_parameters.do_send_supplemental)
 	{
 	  ret = _gnutls_send_supplemental (session, AGAIN (STATE70));
 	  STATE = STATE70;
@@ -3053,7 +2958,7 @@ _gnutls_handshake_server (gnutls_session_t session)
       IMED_RET ("send server hello done", ret, 0);
 
     case STATE71:
-      if (session->security_parameters.extensions.do_recv_supplemental)
+      if (session->security_parameters.do_recv_supplemental)
 	{
 	  ret = _gnutls_recv_supplemental (session);
 	  STATE = STATE71;
@@ -3111,13 +3016,9 @@ _gnutls_handshake_common (gnutls_session_t session)
 	{
 	case STATE0:
 	case STATE40:
-	  if (session->internals.session_ticket_renew)
-	    {
-	      ret =
-		_gnutls_send_new_session_ticket (session, AGAIN (STATE40));
-	      STATE = STATE40;
-	      IMED_RET ("send handshake new session ticket", ret, 0);
-	    }
+          ret = _gnutls_send_new_session_ticket (session, AGAIN (STATE40));
+          STATE = STATE40;
+          IMED_RET ("send handshake new session ticket", ret, 0);
 	  STATE = STATE0;
 	default:
 	  break;
@@ -3145,12 +3046,9 @@ _gnutls_handshake_common (gnutls_session_t session)
 	{
 	case STATE0:
 	case STATE41:
-	  if (session->internals.session_ticket_renew)
-	    {
-	      ret = _gnutls_recv_new_session_ticket (session);
-	      STATE = STATE41;
-	      IMED_RET ("recv handshake new session ticket", ret, 1);
-	    }
+          ret = _gnutls_recv_new_session_ticket (session);
+	  STATE = STATE41;
+	  IMED_RET ("recv handshake new session ticket", ret, 1);
 	  STATE = STATE0;
 	default:
 	  break;
