@@ -4,9 +4,9 @@
  *
  * Author: Nikos Mavrogiannopoulos
  *
- * This file is part of GNUTLS.
+ * This file is part of GnuTLS.
  *
- * The GNUTLS library is free software; you can redistribute it and/or
+ * The GnuTLS is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public License
  * as published by the Free Software Foundation; either version 2.1 of
  * the License, or (at your option) any later version.
@@ -26,8 +26,7 @@
 typedef enum Optional
 { OPTIONAL_PACKET, MANDATORY_PACKET } Optional;
 
-int _gnutls_send_handshake (gnutls_session_t session, void *i_data,
-			    uint32_t i_datasize,
+int _gnutls_send_handshake (gnutls_session_t session, mbuffer_st *bufel,
 			    gnutls_handshake_description_t type);
 int _gnutls_recv_hello_request (gnutls_session_t session, void *data,
 				uint32_t data_size);
@@ -62,4 +61,5 @@ void _gnutls_handshake_hash_buffers_clear (gnutls_session_t session);
 /* This returns true if we have got there
  * before (and not finished due to an interrupt).
  */
-#define AGAIN(target) STATE==target?1:0
+#define AGAIN(target) (STATE==target?1:0)
+#define AGAIN2(state, target) (state==target?1:0)
