@@ -38,6 +38,39 @@
 #include <gnutls_state.h>
 #include <gnutls_datum.h>
 #include <gnutls_rsa_export.h>
+#include <gnutls_mbuffers.h>
+#include "../libextra/ext_inner_application.h" /* isn't this too much? */
+
+/* This is a temporary function to be used before the generate_*
+   internal API is changed to use mbuffers. For now we don't avoid the
+   extra alloc + memcpy. */
+static inline int
+send_handshake (gnutls_session_t session, opaque *data, size_t size,
+		gnutls_handshake_description_t type)
+{
+  mbuffer_st *bufel;
+
+  if(data == NULL && size == 0)
+    return _gnutls_send_handshake(session, NULL, type);
+
+  if (data == NULL && size > 0)
+    {
+      gnutls_assert ();
+      return GNUTLS_E_INVALID_REQUEST;
+    }
+
+  bufel = _gnutls_handshake_alloc(size, size);
+  if (bufel == NULL)
+    {
+      gnutls_assert ();
+      return GNUTLS_E_MEMORY_ERROR;
+    }
+
+  _mbuffer_set_udata(bufel, data, size);
+
+  return _gnutls_send_handshake(session, bufel, type);
+}
+
 
 /* This file contains important thing for the TLS handshake procedure.
  */
@@ -64,15 +97,15 @@ generate_normal_master (gnutls_session_t session, int keep_premaster)
 
   _gnutls_hard_log ("INT: PREMASTER SECRET[%d]: %s\n", PREMASTER.size,
 		    _gnutls_bin2hex (PREMASTER.data, PREMASTER.size, buf,
-				     sizeof (buf)));
+				     sizeof (buf), NULL));
   _gnutls_hard_log ("INT: CLIENT RANDOM[%d]: %s\n", 32,
 		    _gnutls_bin2hex (session->
 				     security_parameters.client_random, 32,
-				     buf, sizeof (buf)));
+				     buf, sizeof (buf), NULL));
   _gnutls_hard_log ("INT: SERVER RANDOM[%d]: %s\n", 32,
 		    _gnutls_bin2hex (session->
 				     security_parameters.server_random, 32,
-				     buf, sizeof (buf)));
+				     buf, sizeof (buf), NULL));
 
   if (gnutls_protocol_get_version (session) == GNUTLS_SSL3)
     {
@@ -91,64 +124,6 @@ generate_normal_master (gnutls_session_t session, int keep_premaster)
 				      security_parameters.master_secret);
 
     }
-  else if (session->security_parameters.extensions.oprfi_client_len > 0 &&
-	   session->security_parameters.extensions.oprfi_server_len > 0)
-    {
-      opaque *rnd;
-      size_t rndlen = 2 * GNUTLS_RANDOM_SIZE;
-
-      rndlen += session->security_parameters.extensions.oprfi_client_len;
-      rndlen += session->security_parameters.extensions.oprfi_server_len;
-
-      rnd = gnutls_malloc (rndlen + 1);
-      if (!rnd)
-	{
-	  gnutls_assert ();
-	  return GNUTLS_E_MEMORY_ERROR;
-	}
-
-      _gnutls_hard_log ("INT: CLIENT OPRFI[%d]: %s\n",
-			session->security_parameters.extensions.
-			oprfi_server_len,
-			_gnutls_bin2hex (session->
-					 security_parameters.extensions.
-					 oprfi_client,
-					 session->
-					 security_parameters.extensions.
-					 oprfi_client_len, buf,
-					 sizeof (buf)));
-      _gnutls_hard_log ("INT: SERVER OPRFI[%d]: %s\n",
-			session->security_parameters.extensions.
-			oprfi_server_len,
-			_gnutls_bin2hex (session->
-					 security_parameters.extensions.
-					 oprfi_server,
-					 session->
-					 security_parameters.extensions.
-					 oprfi_server_len, buf,
-					 sizeof (buf)));
-
-      memcpy (rnd, session->security_parameters.client_random,
-	      GNUTLS_RANDOM_SIZE);
-      memcpy (rnd + GNUTLS_RANDOM_SIZE,
-	      session->security_parameters.extensions.oprfi_client,
-	      session->security_parameters.extensions.oprfi_client_len);
-      memcpy (rnd + GNUTLS_RANDOM_SIZE +
-	      session->security_parameters.extensions.oprfi_client_len,
-	      session->security_parameters.server_random, GNUTLS_RANDOM_SIZE);
-      memcpy (rnd + GNUTLS_RANDOM_SIZE +
-	      session->security_parameters.extensions.oprfi_client_len +
-	      GNUTLS_RANDOM_SIZE,
-	      session->security_parameters.extensions.oprfi_server,
-	      session->security_parameters.extensions.oprfi_server_len);
-
-      ret = _gnutls_PRF (session, PREMASTER.data, PREMASTER.size,
-			 MASTER_SECRET, strlen (MASTER_SECRET),
-			 rnd, rndlen, GNUTLS_MASTER_SIZE,
-			 session->security_parameters.master_secret);
-
-      gnutls_free (rnd);
-    }
   else
     {
       opaque rnd[2 * GNUTLS_RANDOM_SIZE + 1];
@@ -166,8 +141,7 @@ generate_normal_master (gnutls_session_t session, int keep_premaster)
     }
 
   /* TLS/IA inner secret is derived from the master secret. */
-  memcpy (session->security_parameters.inner_secret,
-	  session->security_parameters.master_secret, GNUTLS_MASTER_SIZE);
+  _gnutls_ia_derive_inner_secret(session);
 
   if (!keep_premaster)
     _gnutls_free_datum (&PREMASTER);
@@ -178,7 +152,7 @@ generate_normal_master (gnutls_session_t session, int keep_premaster)
   _gnutls_hard_log ("INT: MASTER SECRET: %s\n",
 		    _gnutls_bin2hex (session->
 				     security_parameters.master_secret,
-				     GNUTLS_MASTER_SIZE, buf, sizeof (buf)));
+				     GNUTLS_MASTER_SIZE, buf, sizeof (buf), NULL));
 
   return ret;
 }
@@ -220,9 +194,8 @@ _gnutls_send_server_kx_message (gnutls_session_t session, int again)
 	}
     }
 
-  ret =
-    _gnutls_send_handshake (session, data, data_size,
-			    GNUTLS_HANDSHAKE_SERVER_KEY_EXCHANGE);
+  ret = send_handshake (session, data, data_size,
+			GNUTLS_HANDSHAKE_SERVER_KEY_EXCHANGE);
   gnutls_free (data);
 
   if (ret < 0)
@@ -266,9 +239,8 @@ _gnutls_send_server_certificate_request (gnutls_session_t session, int again)
 	  return data_size;
 	}
     }
-  ret =
-    _gnutls_send_handshake (session, data, data_size,
-			    GNUTLS_HANDSHAKE_CERTIFICATE_REQUEST);
+  ret = send_handshake (session, data, data_size,
+			GNUTLS_HANDSHAKE_CERTIFICATE_REQUEST);
   gnutls_free (data);
 
   if (ret < 0)
@@ -308,9 +280,8 @@ _gnutls_send_client_kx_message (gnutls_session_t session, int again)
 	  return data_size;
 	}
     }
-  ret =
-    _gnutls_send_handshake (session, data, data_size,
-			    GNUTLS_HANDSHAKE_CLIENT_KEY_EXCHANGE);
+  ret =  send_handshake (session, data, data_size,
+			 GNUTLS_HANDSHAKE_CLIENT_KEY_EXCHANGE);
   gnutls_free (data);
 
   if (ret < 0)
@@ -368,9 +339,8 @@ _gnutls_send_client_certificate_verify (gnutls_session_t session, int again)
 	return 0;
 
     }
-  ret =
-    _gnutls_send_handshake (session, data,
-			    data_size, GNUTLS_HANDSHAKE_CERTIFICATE_VERIFY);
+  ret = send_handshake (session, data, data_size,
+			GNUTLS_HANDSHAKE_CERTIFICATE_VERIFY);
   gnutls_free (data);
 
   return ret;
@@ -550,9 +520,8 @@ _gnutls_send_client_certificate (gnutls_session_t session, int again)
   else
     {				/* TLS 1.0 or SSL 3.0 with a valid certificate 
 				 */
-      ret =
-	_gnutls_send_handshake (session, data, data_size,
-				GNUTLS_HANDSHAKE_CERTIFICATE_PKT);
+      ret = send_handshake (session, data, data_size,
+			    GNUTLS_HANDSHAKE_CERTIFICATE_PKT);
       gnutls_free (data);
     }
 
@@ -595,9 +564,8 @@ _gnutls_send_server_certificate (gnutls_session_t session, int again)
 	  return data_size;
 	}
     }
-  ret =
-    _gnutls_send_handshake (session, data, data_size,
-			    GNUTLS_HANDSHAKE_CERTIFICATE_PKT);
+  ret = send_handshake (session, data, data_size,
+			GNUTLS_HANDSHAKE_CERTIFICATE_PKT);
   gnutls_free (data);
 
   if (ret < 0)
