@@ -32,6 +32,7 @@
 #include "gnutls_compress.h"
 #include "gnutls_cipher.h"
 #include "gnutls_buffers.h"
+#include "gnutls_mbuffers.h"
 #include "gnutls_handshake.h"
 #include "gnutls_hash_int.h"
 #include "gnutls_cipher_int.h"
@@ -272,7 +273,7 @@ gnutls_bye (gnutls_session_t session, gnutls_close_request_t how)
 inline static void
 session_invalidate (gnutls_session_t session)
 {
-  session->internals.valid_connection = VALID_FALSE;
+  session->internals.invalid_connection = 1;
 }
 
 
@@ -287,7 +288,7 @@ session_unresumable (gnutls_session_t session)
 inline static int
 session_is_valid (gnutls_session_t session)
 {
-  if (session->internals.valid_connection == VALID_FALSE)
+  if (session->internals.invalid_connection != 0)
     return GNUTLS_E_INVALID_SESSION;
 
   return 0;
@@ -334,10 +335,10 @@ copy_record_version (gnutls_session_t session,
 ssize_t
 _gnutls_send_int (gnutls_session_t session, content_type_t type,
 		  gnutls_handshake_description_t htype, const void *_data,
-		  size_t sizeofdata)
+		  size_t sizeofdata, unsigned int mflags)
 {
-  uint8_t *cipher;
-  int cipher_size;
+  mbuffer_st *bufel;
+  size_t cipher_size;
   int retval, ret;
   int data2send_size;
   uint8_t headers[5];
@@ -347,7 +348,7 @@ _gnutls_send_int (gnutls_session_t session, content_type_t type,
    * If the previous send was interrupted then a null pointer is
    * ok, and means to resume.
    */
-  if (session->internals.record_send_buffer.length == 0 &&
+  if (session->internals.record_send_buffer.byte_length == 0 &&
       (sizeofdata == 0 && _data == NULL))
     {
       gnutls_assert ();
@@ -383,7 +384,7 @@ _gnutls_send_int (gnutls_session_t session, content_type_t type,
   /* Only encrypt if we don't have data to send 
    * from the previous run. - probably interrupted.
    */
-  if (session->internals.record_send_buffer.length > 0)
+  if (mflags != 0 && session->internals.record_send_buffer.byte_length > 0)
     {
       ret = _gnutls_io_write_flush (session);
       if (ret > 0)
@@ -391,9 +392,7 @@ _gnutls_send_int (gnutls_session_t session, content_type_t type,
       else
 	cipher_size = 0;
 
-      cipher = NULL;
-
-      retval = session->internals.record_send_buffer_user_size;
+      retval = session->internals.record_send_buffer.byte_length;
     }
   else
     {
@@ -401,8 +400,8 @@ _gnutls_send_int (gnutls_session_t session, content_type_t type,
       /* now proceed to packet encryption
        */
       cipher_size = data2send_size + MAX_RECORD_OVERHEAD;
-      cipher = gnutls_malloc (cipher_size);
-      if (cipher == NULL)
+      bufel = _mbuffer_alloc (cipher_size, cipher_size);
+      if (bufel == NULL)
 	{
 	  gnutls_assert ();
 	  return GNUTLS_E_MEMORY_ERROR;
@@ -410,7 +409,7 @@ _gnutls_send_int (gnutls_session_t session, content_type_t type,
 
       cipher_size =
 	_gnutls_encrypt (session, headers, RECORD_HEADER_SIZE, data,
-			 data2send_size, cipher, cipher_size, type,
+			 data2send_size, _mbuffer_get_udata_ptr(bufel), cipher_size, type,
 			 (session->internals.priorities.no_padding ==
 			  0) ? 1 : 0);
       if (cipher_size <= 0)
@@ -418,12 +417,11 @@ _gnutls_send_int (gnutls_session_t session, content_type_t type,
 	  gnutls_assert ();
 	  if (cipher_size == 0)
 	    cipher_size = GNUTLS_E_ENCRYPTION_FAILED;
-	  gnutls_free (cipher);
+	  gnutls_free (bufel);
 	  return cipher_size;	/* error */
 	}
 
       retval = data2send_size;
-      session->internals.record_send_buffer_user_size = data2send_size;
 
       /* increase sequence number
        */
@@ -432,12 +430,12 @@ _gnutls_send_int (gnutls_session_t session, content_type_t type,
 	{
 	  session_invalidate (session);
 	  gnutls_assert ();
-	  gnutls_free (cipher);
+	  gnutls_free (bufel);
 	  return GNUTLS_E_RECORD_LIMIT_REACHED;
 	}
 
-      ret = _gnutls_io_write_buffered (session, cipher, cipher_size);
-      gnutls_free (cipher);
+      _mbuffer_set_udata_size(bufel, cipher_size);
+      ret = _gnutls_io_write_buffered (session, bufel, mflags);
     }
 
   if (ret != cipher_size)
@@ -462,14 +460,12 @@ _gnutls_send_int (gnutls_session_t session, content_type_t type,
       return ret;
     }
 
-  session->internals.record_send_buffer_user_size = 0;
-
   _gnutls_record_log ("REC[%p]: Sent Packet[%d] %s(%d) with length: %d\n",
 		      session,
 		      (int)
 		      _gnutls_uint64touint32
 		      (&session->connection_state.write_sequence_number),
-		      _gnutls_packet2str (type), type, cipher_size);
+		      _gnutls_packet2str (type), type, (int)cipher_size);
 
   return retval;
 }
@@ -485,7 +481,7 @@ _gnutls_send_change_cipher_spec (gnutls_session_t session, int again)
   _gnutls_handshake_log ("REC[%p]: Sent ChangeCipherSpec\n", session);
 
   if (again == 0)
-    return _gnutls_send_int (session, GNUTLS_CHANGE_CIPHER_SPEC, -1, data, 1);
+    return _gnutls_send_int (session, GNUTLS_CHANGE_CIPHER_SPEC, -1, data, 1, MBUFFER_FLUSH);
   else
     {
       return _gnutls_io_write_flush (session);
@@ -744,14 +740,12 @@ record_check_type (gnutls_session_t session,
 	  if (session->security_parameters.entity == GNUTLS_SERVER)
 	    {
 	      gnutls_assert ();
-	      ret =
-		_gnutls_record_buffer_put (recv_type, session, (void *) data,
-					   data_size);
-	      if (ret < 0)
-		{
-		  gnutls_assert ();
-		  return ret;
-		}
+	      ret = _gnutls_record_buffer_put (recv_type, session, (void *) data, data_size);
+	      if (ret < 0) 
+	        {
+	          gnutls_assert();
+	          return ret;
+                }
 	      return GNUTLS_E_REHANDSHAKE;
 	    }
 
@@ -801,8 +795,7 @@ get_temp_recv_buffer (gnutls_session_t session, gnutls_datum_t * tmp)
 {
   size_t max_record_size;
 
-  if (gnutls_compression_get (session) != GNUTLS_COMP_NULL ||
-      session->internals.priorities.allow_large_records != 0)
+  if (gnutls_compression_get (session) != GNUTLS_COMP_NULL)
     max_record_size = MAX_RECORD_RECV_SIZE + EXTRA_COMP_SIZE;
   else
     max_record_size = MAX_RECORD_RECV_SIZE;
@@ -1140,7 +1133,7 @@ gnutls_record_send (gnutls_session_t session, const void *data,
 		    size_t sizeofdata)
 {
   return _gnutls_send_int (session, GNUTLS_APPLICATION_DATA, -1, data,
-			   sizeofdata);
+			   sizeofdata, MBUFFER_FLUSH);
 }
 
 /**
@@ -1180,64 +1173,3 @@ gnutls_record_recv (gnutls_session_t session, void *data, size_t sizeofdata)
 			   sizeofdata);
 }
 
-/**
- * gnutls_record_get_max_size:
- * @session: is a #gnutls_session_t structure.
- *
- * Get the record size.  The maximum record size is negotiated by the
- * client after the first handshake message.
- *
- * Returns: The maximum record packet size in this connection.
- **/
-size_t
-gnutls_record_get_max_size (gnutls_session_t session)
-{
-  /* Recv will hold the negotiated max record size
-   * always.
-   */
-  return session->security_parameters.max_record_recv_size;
-}
-
-
-/**
- * gnutls_record_set_max_size:
- * @session: is a #gnutls_session_t structure.
- * @size: is the new size
- *
- * This function sets the maximum record packet size in this
- * connection.  This property can only be set to clients.  The server
- * may choose not to accept the requested size.
- *
- * Acceptable values are 512(=2^9), 1024(=2^10), 2048(=2^11) and
- * 4096(=2^12).  The requested record size does get in effect
- * immediately only while sending data. The receive part will take
- * effect after a successful handshake.
- *
- * This function uses a TLS extension called 'max record size'.  Not
- * all TLS implementations use or even understand this extension.
- *
- * Returns: On success, %GNUTLS_E_SUCCESS (zero) is returned,
- *   otherwise an error code is returned.
- **/
-ssize_t
-gnutls_record_set_max_size (gnutls_session_t session, size_t size)
-{
-  ssize_t new_size;
-
-  if (session->security_parameters.entity == GNUTLS_SERVER)
-    return GNUTLS_E_INVALID_REQUEST;
-
-  new_size = _gnutls_mre_record2num (size);
-
-  if (new_size < 0)
-    {
-      gnutls_assert ();
-      return new_size;
-    }
-
-  session->security_parameters.max_record_send_size = size;
-
-  session->internals.proposed_record_size = size;
-
-  return 0;
-}
