@@ -26,14 +26,16 @@
 #ifndef GNUTLS_SIG_H
 # define GNUTLS_SIG_H
 
+#include <gnutls/abstract.h>
+
 int _gnutls_handshake_sign_cert_vrfy (gnutls_session_t session,
 				      gnutls_cert * cert,
-				      gnutls_privkey * pkey,
+				      gnutls_privkey_t pkey,
 				      gnutls_datum_t * signature);
 
 int _gnutls_handshake_sign_data (gnutls_session_t session,
 				 gnutls_cert * cert,
-				 gnutls_privkey * pkey,
+				 gnutls_privkey_t  pkey,
 				 gnutls_datum_t * params,
 				 gnutls_datum_t * signature,
 				 gnutls_sign_algorithm_t * algo);
@@ -49,7 +51,7 @@ int _gnutls_handshake_verify_data (gnutls_session_t session,
 				   gnutls_datum_t * signature,
 				   gnutls_sign_algorithm_t algo);
 
-int _gnutls_sign (gnutls_pk_algorithm_t algo,
+int _gnutls_soft_sign (gnutls_pk_algorithm_t algo,
 		  bigint_t * params, int params_size,
 		  const gnutls_datum_t * data, gnutls_datum_t * signature);
 
