@@ -38,19 +38,19 @@
 #include <gnutls_handshake.h>
 #include <gnutls_dh.h>
 #include <gnutls_buffers.h>
+#include <gnutls_mbuffers.h>
 #include <gnutls_state.h>
 #include <auth_cert.h>
 #include <auth_anon.h>
 #include <auth_psk.h>
 #include <gnutls_algorithms.h>
 #include <gnutls_rsa_export.h>
+#include <gnutls_extensions.h>
+#include <system.h>
 
 /* These should really be static, but src/tests.c calls them.  Make
    them public functions?  */
 void
-_gnutls_record_set_default_version (gnutls_session_t session,
-				    unsigned char major, unsigned char minor);
-void
 _gnutls_rsa_pms_set_version (gnutls_session_t session,
 			     unsigned char major, unsigned char minor);
 
@@ -157,7 +157,7 @@ _gnutls_session_cert_type_supported (gnutls_session_t session,
       if (cred == NULL)
 	return GNUTLS_E_UNSUPPORTED_CERTIFICATE_TYPE;
 
-      if (cred->server_get_cert_callback == NULL)
+      if (cred->server_get_cert_callback == NULL && cred->get_cert_callback == NULL)
 	{
 	  for (i = 0; i < cred->ncerts; i++)
 	    {
@@ -218,7 +218,6 @@ _gnutls_handshake_internal_state_init (gnutls_session_t session)
   session->internals.extensions_sent_size = 0;
 
   /* by default no selected certificate */
-  session->internals.proposed_record_size = DEFAULT_MAX_RECORD_SIZE;
   session->internals.adv_version_major = 0;
   session->internals.adv_version_minor = 0;
   session->internals.v2_hello = 0;
@@ -227,7 +226,6 @@ _gnutls_handshake_internal_state_init (gnutls_session_t session)
   session->internals.adv_version_minor = 0;
   session->internals.adv_version_minor = 0;
   session->internals.direction = 0;
-  session->internals.safe_renegotiation_received = 0;
 
   /* use out of band data for the last
    * handshake messages received.
@@ -297,10 +295,10 @@ gnutls_init (gnutls_session_t * session, gnutls_connection_end_t con_end)
   _gnutls_buffer_init (&(*session)->internals.handshake_hash_buffer);
   _gnutls_buffer_init (&(*session)->internals.ia_data_buffer);
 
-  _gnutls_buffer_init (&(*session)->internals.record_send_buffer);
+  _mbuffer_init (&(*session)->internals.record_send_buffer);
   _gnutls_buffer_init (&(*session)->internals.record_recv_buffer);
 
-  _gnutls_buffer_init (&(*session)->internals.handshake_send_buffer);
+  _mbuffer_init (&(*session)->internals.handshake_send_buffer);
   _gnutls_buffer_init (&(*session)->internals.handshake_recv_buffer);
 
   (*session)->key = gnutls_calloc (1, sizeof (struct gnutls_key_st));
@@ -353,7 +351,15 @@ gnutls_init (gnutls_session_t * session, gnutls_connection_end_t con_end)
   /* emulate old gnutls behavior for old applications that do not use the priority_*
    * functions.
    */
-  (*session)->internals.priorities.safe_renegotiation = SR_PARTIAL;
+  (*session)->internals.priorities.sr = SR_PARTIAL;
+
+#ifdef HAVE_WRITEV
+  gnutls_transport_set_push_function2(*session, system_writev);
+#else
+  gnutls_transport_set_push_function(*session, system_write);
+#endif
+  gnutls_transport_set_pull_function(*session, system_read);
+  gnutls_transport_set_errno_function(*session, system_errno);
 
   return 0;
 }
@@ -387,6 +393,7 @@ gnutls_deinit (gnutls_session_t session)
 
   _gnutls_handshake_internal_state_clear (session);
   _gnutls_handshake_io_buffer_clear (session);
+  _gnutls_ext_free_session_data (session);
 
   _gnutls_free_datum (&session->connection_state.read_mac_secret);
   _gnutls_free_datum (&session->connection_state.write_mac_secret);
@@ -396,7 +403,7 @@ gnutls_deinit (gnutls_session_t session)
   _gnutls_buffer_clear (&session->internals.handshake_data_buffer);
   _gnutls_buffer_clear (&session->internals.application_data_buffer);
   _gnutls_buffer_clear (&session->internals.record_recv_buffer);
-  _gnutls_buffer_clear (&session->internals.record_send_buffer);
+  _mbuffer_clear (&session->internals.record_send_buffer);
 
   gnutls_credentials_clear (session);
   _gnutls_selected_certs_deinit (session);
@@ -441,19 +448,6 @@ gnutls_deinit (gnutls_session_t session)
       session->key = NULL;
     }
 
-  gnutls_free (session->internals.srp_username);
-
-  if (session->internals.srp_password)
-    {
-      memset (session->internals.srp_password, 0,
-	      strlen (session->internals.srp_password));
-      gnutls_free (session->internals.srp_password);
-    }
-
-  gnutls_free (session->security_parameters.extensions.session_ticket);
-  gnutls_free (session->internals.resumed_security_parameters.
-	       extensions.session_ticket);
-
   memset (session, 0, sizeof (struct gnutls_session_int));
   gnutls_free (session);
 }
