@@ -3,9 +3,9 @@
  *
  * Author: Nikos Mavrogiannopoulos
  *
- * This file is part of GNUTLS.
+ * This file is part of GnuTLS.
  *
- * The GNUTLS library is free software; you can redistribute it and/or
+ * The GnuTLS is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public License
  * as published by the Free Software Foundation; either version 2.1 of
  * the License, or (at your option) any later version.
@@ -35,6 +35,11 @@ int gnutls_cipher_encrypt (const gnutls_cipher_hd_t handle,
 			   void *text, size_t textlen);
 int gnutls_cipher_decrypt (const gnutls_cipher_hd_t handle,
 			   void *ciphertext, size_t ciphertextlen);
+int gnutls_cipher_decrypt2 (gnutls_cipher_hd_t handle, const void *ciphertext,
+		       size_t ciphertextlen, void* text, size_t textlen);
+int gnutls_cipher_encrypt2 (gnutls_cipher_hd_t handle, void *text, size_t textlen,
+  void* ciphertext, size_t ciphertextlen);
+
 void gnutls_cipher_deinit (gnutls_cipher_hd_t handle);
 int gnutls_cipher_get_block_size (gnutls_cipher_algorithm_t algorithm);
 
@@ -112,6 +117,9 @@ typedef enum gnutls_rnd_level
   GNUTLS_RND_KEY = 2
 } gnutls_rnd_level_t;
 
+int gnutls_rnd (gnutls_rnd_level_t level, void *data, size_t len);
+
+
 /**
  * gnutls_pk_flag_t:
  * @GNUTLS_PK_FLAG_NONE: No flag.
@@ -214,7 +222,7 @@ typedef struct gnutls_crypto_bigint
 		       gnutls_bigint_format_t format);
 } gnutls_crypto_bigint_st;
 
-#define GNUTLS_MAX_PK_PARAMS 6
+#define GNUTLS_MAX_PK_PARAMS 16
 
 typedef struct
 {
@@ -234,9 +242,11 @@ void gnutls_pk_params_init (gnutls_pk_params_st * p);
  *  [3] is prime1 (p) (private key only)
  *  [4] is prime2 (q) (private key only)
  *  [5] is coefficient (u == inverse of p mod q) (private key only)
+ *  [6] e1 == d mod (p-1)
+ *  [7] e2 == d mod (q-1)
  *
- *  note that other packages use inverse of q mod p,
- *  so we need to perform conversions using fixup_params().
+ *  note that for libgcrypt that does not use the inverse of q mod p,
+ *  we need to perform conversions using fixup_params().
  *
  * DSA:
  *  [0] is p
