@@ -78,7 +78,6 @@ extern "C"
 #define GNUTLS_KP_EMAIL_PROTECTION		"1.3.6.1.5.5.7.3.4"
 #define GNUTLS_KP_TIME_STAMPING			"1.3.6.1.5.5.7.3.8"
 #define GNUTLS_KP_OCSP_SIGNING			"1.3.6.1.5.5.7.3.9"
-#define GNUTLS_KP_IPSEC_IKE			"1.3.6.1.5.5.7.3.17"
 #define GNUTLS_KP_ANY				"2.5.29.37.0"
 
 #define GNUTLS_FSAN_SET 0
@@ -140,6 +139,8 @@ extern "C"
 				  unsigned int flags,
 				  unsigned char *output_data,
 				  size_t * output_data_size);
+  int gnutls_x509_crt_get_preferred_hash_algorithm (gnutls_x509_crt_t crt,
+				      gnutls_digest_algorithm_t * hash, unsigned int *mand);
 
   int gnutls_x509_crt_set_authority_key_id (gnutls_x509_crt_t cert,
 					    const void *id, size_t id_size);
@@ -506,10 +507,10 @@ extern "C"
  * @GNUTLS_VERIFY_DISABLE_TRUSTED_TIME_CHECKS: If set a signer in the trusted
  *   list is never checked for expiration or activation.
  * @GNUTLS_VERIFY_ALLOW_X509_V1_CA_CRT: Allow only trusted CA
- *   certificates that have version 1.  This is the default.
- * @GNUTLS_VERIFY_DO_NOT_ALLOW_X509_V1_CA_CRT: Do not allow trusted CA
- *   certificates that have version 1.  This option is to be used
- *   to deprecate all V1 certificates.
+ *   certificates that have version 1.  This is safer than
+ *   %GNUTLS_VERIFY_ALLOW_ANY_X509_V1_CA_CRT, and should be used
+ *   instead. That way only signers in your trusted list will be
+ *   allowed to have certificates of version 1.
  * @GNUTLS_VERIFY_DO_NOT_ALLOW_SAME: If a certificate is not signed by
  *   anyone trusted but exists in the trusted CA list do not treat it
  *   as trusted.
@@ -538,7 +539,6 @@ extern "C"
     GNUTLS_VERIFY_ALLOW_SIGN_RSA_MD5 = 32,
     GNUTLS_VERIFY_DISABLE_TIME_CHECKS = 64,
     GNUTLS_VERIFY_DISABLE_TRUSTED_TIME_CHECKS = 128,
-    GNUTLS_VERIFY_DO_NOT_ALLOW_X509_V1_CA_CRT = 256
   } gnutls_certificate_verify_flags;
 
   int gnutls_x509_crt_check_issuer (gnutls_x509_crt_t cert,
@@ -613,7 +613,7 @@ extern "C"
     GNUTLS_PKCS_USE_PBES2_3DES = 16,
     GNUTLS_PKCS_USE_PBES2_AES_128 = 32,
     GNUTLS_PKCS_USE_PBES2_AES_192 = 64,
-    GNUTLS_PKCS_USE_PBES2_AES_256 = 128
+    GNUTLS_PKCS_USE_PBES2_AES_256 = 128,
   } gnutls_pkcs_encrypt_flags_t;
 
   int gnutls_x509_privkey_init (gnutls_x509_privkey_t * key);
@@ -635,6 +635,15 @@ extern "C"
 					  const gnutls_datum_t * p,
 					  const gnutls_datum_t * q,
 					  const gnutls_datum_t * u);
+  int gnutls_x509_privkey_import_rsa_raw2 (gnutls_x509_privkey_t key,
+					  const gnutls_datum_t * m,
+					  const gnutls_datum_t * e,
+					  const gnutls_datum_t * d,
+					  const gnutls_datum_t * p,
+					  const gnutls_datum_t * q,
+					  const gnutls_datum_t * u,
+					  const gnutls_datum_t *exp1,
+					  const gnutls_datum_t *exp2);
   int gnutls_x509_privkey_fix (gnutls_x509_privkey_t key);
 
   int gnutls_x509_privkey_export_dsa_raw (gnutls_x509_privkey_t key,
@@ -670,6 +679,11 @@ extern "C"
 					unsigned int flags,
 					void *output_data,
 					size_t * output_data_size);
+  int gnutls_x509_privkey_export_rsa_raw2 (gnutls_x509_privkey_t key,
+        gnutls_datum_t * m, gnutls_datum_t * e,
+        gnutls_datum_t * d, gnutls_datum_t * p,
+        gnutls_datum_t * q, gnutls_datum_t * u,
+        gnutls_datum_t* e1, gnutls_datum_t* e2);
   int gnutls_x509_privkey_export_rsa_raw (gnutls_x509_privkey_t key,
 					  gnutls_datum_t * m,
 					  gnutls_datum_t * e,
@@ -686,6 +700,11 @@ extern "C"
 				     const gnutls_datum_t * data,
 				     void *signature,
 				     size_t * signature_size);
+  int gnutls_x509_privkey_sign_data2 (gnutls_x509_privkey_t key,
+				     gnutls_digest_algorithm_t digest,
+				     unsigned int flags,
+				     const gnutls_datum_t * data,
+				     gnutls_datum_t *signature);
   int gnutls_x509_privkey_verify_data (gnutls_x509_privkey_t key,
 				       unsigned int flags,
 				       const gnutls_datum_t * data,
@@ -722,6 +741,9 @@ extern "C"
   int gnutls_x509_crq_import (gnutls_x509_crq_t crq,
 			      const gnutls_datum_t * data,
 			      gnutls_x509_crt_fmt_t format);
+  int gnutls_x509_crq_get_preferred_hash_algorithm (gnutls_x509_crq_t crq,
+			      gnutls_digest_algorithm_t * hash, unsigned int* mand);
+
   int gnutls_x509_crq_get_dn (gnutls_x509_crq_t crq, char *buf,
 			      size_t * sizeof_buf);
   int gnutls_x509_crq_get_dn_oid (gnutls_x509_crq_t crq, int indx,
