@@ -91,7 +91,7 @@ _gnutls_x509_parse_dn (ASN1_TYPE asn1_struct,
 		       const char *asn1_rdn_name, char *buf,
 		       size_t * sizeof_buf)
 {
-  gnutls_string out_str;
+  gnutls_buffer_st out_str;
   int k2, k1, result;
   char tmpbuffer1[ASN1_MAX_NAME_SIZE];
   char tmpbuffer2[ASN1_MAX_NAME_SIZE];
@@ -115,7 +115,7 @@ _gnutls_x509_parse_dn (ASN1_TYPE asn1_struct,
   else
     *sizeof_buf = 0;
 
-  _gnutls_string_init (&out_str, gnutls_malloc, gnutls_realloc, gnutls_free);
+  _gnutls_buffer_init (&out_str);
 
   k1 = 0;
   do
@@ -214,7 +214,7 @@ _gnutls_x509_parse_dn (ASN1_TYPE asn1_struct,
 	      result = _gnutls_asn2err (result);
 	      goto cleanup;
 	    }
-#define STR_APPEND(y) if ((result=_gnutls_string_append_str( &out_str, y)) < 0) { \
+#define STR_APPEND(y) if ((result=_gnutls_buffer_append_str( &out_str, y)) < 0) { \
 	gnutls_assert(); \
 	goto cleanup; \
 }
@@ -281,7 +281,7 @@ _gnutls_x509_parse_dn (ASN1_TYPE asn1_struct,
 	      gnutls_assert ();
 	      _gnutls_x509_log
 		("Found OID: '%s' with value '%s'\n",
-		 oid, _gnutls_bin2hex (value2, len, escaped, sizeof_escaped));
+		 oid, _gnutls_bin2hex (value2, len, escaped, sizeof_escaped, NULL));
 	      goto cleanup;
 	    }
 	  STR_APPEND (str_escape (string, escaped, sizeof_escaped));
@@ -309,7 +309,7 @@ _gnutls_x509_parse_dn (ASN1_TYPE asn1_struct,
 
   if (buf)
     {
-      _gnutls_string_get_data (&out_str, buf, sizeof_buf);
+      _gnutls_buffer_pop_data (&out_str, buf, sizeof_buf);
       buf[*sizeof_buf] = 0;
     }
   else
@@ -321,7 +321,7 @@ cleanup:
   gnutls_free (value2);
   gnutls_free (string);
   gnutls_free (escaped);
-  _gnutls_string_clear (&out_str);
+  _gnutls_buffer_clear (&out_str);
   return result;
 }
 
@@ -655,7 +655,7 @@ _gnutls_x509_encode_and_write_attribute (const char *given_oid,
 
   /* Find how to encode the data.
    */
-  val_name = asn1_find_structure_from_oid (_gnutls_get_pkix (), given_oid);
+  val_name = _gnutls_x509_oid2asn_string(given_oid);
   if (val_name == NULL)
     {
       gnutls_assert ();
@@ -663,10 +663,7 @@ _gnutls_x509_encode_and_write_attribute (const char *given_oid,
       return GNUTLS_E_X509_UNSUPPORTED_OID;
     }
 
-  _gnutls_str_cpy (tmp, sizeof (tmp), "PKIX1.");
-  _gnutls_str_cat (tmp, sizeof (tmp), val_name);
-
-  result = asn1_create_element (_gnutls_get_pkix (), tmp, &c2);
+  result = asn1_create_element (_gnutls_get_pkix (), val_name, &c2);
   if (result != ASN1_SUCCESS)
     {
       gnutls_assert ();
