@@ -41,13 +41,15 @@
 #include <gnutls_datum.h>
 #include <x509_int.h>
 #include <common.h>
+#include <sign.h>
+#include <gnutls/abstract.h>
 
 /* Writes the digest information and the digest in a DER encoded
  * structure. The digest info is allocated and stored into the info structure.
  */
 static int
 encode_ber_digest_info (gnutls_digest_algorithm_t hash,
-			const gnutls_datum_t * digest, gnutls_datum_t * info)
+			const gnutls_datum_t * digest, gnutls_datum_t * output)
 {
   ASN1_TYPE dinfo = ASN1_TYPE_EMPTY;
   int result;
@@ -83,7 +85,7 @@ encode_ber_digest_info (gnutls_digest_algorithm_t hash,
      Regardless of what is correct, this appears to be what most
      implementations do.  */
   result = asn1_write_value (dinfo, "digestAlgorithm.parameters",
-			     ASN1_NULL, ASN1_NULL_SIZE);
+			     "\x05\x00", 2);
   if (result != ASN1_SUCCESS)
     {
       gnutls_assert ();
@@ -99,18 +101,18 @@ encode_ber_digest_info (gnutls_digest_algorithm_t hash,
       return _gnutls_asn2err (result);
     }
 
-  info->size = 0;
-  asn1_der_coding (dinfo, "", NULL, &info->size, NULL);
+  output->size = 0;
+  asn1_der_coding (dinfo, "", NULL, &output->size, NULL);
 
-  info->data = gnutls_malloc (info->size);
-  if (info->data == NULL)
+  output->data = gnutls_malloc (output->size);
+  if (output->data == NULL)
     {
       gnutls_assert ();
       asn1_delete_structure (&dinfo);
       return GNUTLS_E_MEMORY_ERROR;
     }
 
-  result = asn1_der_coding (dinfo, "", info->data, &info->size, NULL);
+  result = asn1_der_coding (dinfo, "", output->data, &output->size, NULL);
   if (result != ASN1_SUCCESS)
     {
       gnutls_assert ();
@@ -128,14 +130,13 @@ encode_ber_digest_info (gnutls_digest_algorithm_t hash,
  * params[0] is modulus
  * params[1] is public key
  */
-static int
-pkcs1_rsa_sign (gnutls_digest_algorithm_t hash, const gnutls_datum_t * text,
-		bigint_t * params, int params_len, gnutls_datum_t * signature)
+int
+pk_pkcs1_rsa_hash (gnutls_digest_algorithm_t hash, const gnutls_datum_t * text, gnutls_datum_t * output)
 {
   int ret;
   opaque _digest[MAX_HASH_SIZE];
   digest_hd_st hd;
-  gnutls_datum_t digest, info;
+  gnutls_datum_t digest;
 
   ret = _gnutls_hash_init (&hd, HASH2MAC (hash));
   if (ret < 0)
@@ -152,104 +153,53 @@ pkcs1_rsa_sign (gnutls_digest_algorithm_t hash, const gnutls_datum_t * text,
 
   /* Encode the digest as a DigestInfo
    */
-  if ((ret = encode_ber_digest_info (hash, &digest, &info)) != 0)
+  if ((ret = encode_ber_digest_info (hash, &digest, output)) != 0)
     {
       gnutls_assert ();
       return ret;
     }
 
-  if ((ret =
-       _gnutls_sign (GNUTLS_PK_RSA, params, params_len, &info,
-		     signature)) < 0)
-    {
-      gnutls_assert ();
-      _gnutls_free_datum (&info);
-      return ret;
-    }
-
-  _gnutls_free_datum (&info);
-
   return 0;
 }
 
-static int
-dsa_sign (const gnutls_datum_t * text,
-	  bigint_t * params, int params_len, gnutls_datum_t * signature)
+int
+pk_dsa_hash (gnutls_digest_algorithm_t hash, const gnutls_datum_t * text, gnutls_datum_t * digest)
 {
   int ret;
-  opaque _digest[MAX_HASH_SIZE];
   digest_hd_st hd;
-  gnutls_datum_t digest;
-  gnutls_digest_algorithm_t hash = _gnutls_dsa_q_to_hash (params[1]);
+
+  if (hash != GNUTLS_DIG_SHA1 && hash != GNUTLS_DIG_SHA224 &&
+    hash != GNUTLS_DIG_SHA256)
+    {
+      gnutls_assert();
+      return GNUTLS_E_INVALID_REQUEST;
+    }
+
+  digest->size = _gnutls_hash_get_algo_len(hash);
+  digest->data = gnutls_malloc( digest->size);
+  if (digest->data == NULL)
+    {
+      gnutls_assert();
+      return GNUTLS_E_MEMORY_ERROR;
+    }
 
   ret = _gnutls_hash_init (&hd, hash);
   if (ret < 0)
     {
       gnutls_assert ();
-      return ret;
+      goto fail;
     }
 
   _gnutls_hash (&hd, text->data, text->size);
-  _gnutls_hash_deinit (&hd, _digest);
 
-  digest.data = _digest;
-  digest.size = _gnutls_hash_get_algo_len (hash);
-
-  if ((ret =
-       _gnutls_sign (GNUTLS_PK_DSA, params, params_len, &digest,
-		     signature)) < 0)
-    {
-      gnutls_assert ();
-      return ret;
-    }
+  _gnutls_hash_deinit (&hd, digest->data);
 
   return 0;
-}
 
-/* Signs the given data using the parameters from the signer's
- * private key.
- *
- * returns 0 on success.
- * 
- * 'tbs' is the data to be signed
- * 'signature' will hold the signature!
- * 'hash' is only used in PKCS1 RSA signing.
- */
-int
-_gnutls_x509_sign (const gnutls_datum_t * tbs,
-		   gnutls_digest_algorithm_t hash,
-		   gnutls_x509_privkey_t signer, gnutls_datum_t * signature)
-{
-  int ret;
-
-  switch (signer->pk_algorithm)
-    {
-    case GNUTLS_PK_RSA:
-      ret =
-	pkcs1_rsa_sign (hash, tbs, signer->params, signer->params_size,
-			signature);
-      if (ret < 0)
-	{
-	  gnutls_assert ();
-	  return ret;
-	}
-      return 0;
-      break;
-    case GNUTLS_PK_DSA:
-      ret = dsa_sign (tbs, signer->params, signer->params_size, signature);
-      if (ret < 0)
-	{
-	  gnutls_assert ();
-	  return ret;
-	}
-
-      return 0;
-      break;
-    default:
-      gnutls_assert ();
-      return GNUTLS_E_INTERNAL_ERROR;
-    }
+fail:
+  gnutls_free(digest->data);
 
+  return ret;
 }
 
 /* This is the same as the _gnutls_x509_sign, but this one will decode
@@ -257,15 +207,12 @@ _gnutls_x509_sign (const gnutls_datum_t * tbs,
  * of the TBS and sign it on the fly.
  */
 int
-_gnutls_x509_sign_tbs (ASN1_TYPE cert, const char *tbs_name,
-		       gnutls_digest_algorithm_t hash,
-		       gnutls_x509_privkey_t signer,
-		       gnutls_datum_t * signature)
+_gnutls_x509_get_tbs (ASN1_TYPE cert, const char *tbs_name,
+		       gnutls_datum_t * tbs)
 {
   int result;
   opaque *buf;
   int buf_size;
-  gnutls_datum_t tbs;
 
   buf_size = 0;
   asn1_der_coding (cert, tbs_name, NULL, &buf_size, NULL);
@@ -286,13 +233,10 @@ _gnutls_x509_sign_tbs (ASN1_TYPE cert, const char *tbs_name,
       return _gnutls_asn2err (result);
     }
 
-  tbs.data = buf;
-  tbs.size = buf_size;
+  tbs->data = buf;
+  tbs->size = buf_size;
 
-  result = _gnutls_x509_sign (&tbs, hash, signer, signature);
-  gnutls_free (buf);
-
-  return result;
+  return 0;
 }
 
 /*-
@@ -311,10 +255,11 @@ int
 _gnutls_x509_pkix_sign (ASN1_TYPE src, const char *src_name,
 			gnutls_digest_algorithm_t dig,
 			gnutls_x509_crt_t issuer,
-			gnutls_x509_privkey_t issuer_key)
+			gnutls_privkey_t issuer_key)
 {
   int result;
   gnutls_datum_t signature;
+  gnutls_datum_t tbs;
   char name[128];
 
   /* Step 1. Copy the issuer's name into the certificate.
@@ -335,9 +280,7 @@ _gnutls_x509_pkix_sign (ASN1_TYPE src, const char *src_name,
   _gnutls_str_cat (name, sizeof (name), ".signature");
 
   result = _gnutls_x509_write_sig_params (src, name,
-					  issuer_key->pk_algorithm, dig,
-					  issuer_key->params,
-					  issuer_key->params_size);
+					  gnutls_privkey_get_pk_algorithm(issuer_key, NULL), dig);
   if (result < 0)
     {
       gnutls_assert ();
@@ -346,7 +289,16 @@ _gnutls_x509_pkix_sign (ASN1_TYPE src, const char *src_name,
 
   /* Step 2. Sign the certificate.
    */
-  result = _gnutls_x509_sign_tbs (src, src_name, dig, issuer_key, &signature);
+  result = _gnutls_x509_get_tbs (src, src_name, &tbs);
+
+  if (result < 0)
+    {
+      gnutls_assert ();
+      return result;
+    }
+
+  result = gnutls_privkey_sign_data (issuer_key, dig, 0, &tbs, &signature);
+  gnutls_free(tbs.data);
 
   if (result < 0)
     {
@@ -372,9 +324,7 @@ _gnutls_x509_pkix_sign (ASN1_TYPE src, const char *src_name,
    */
 
   result = _gnutls_x509_write_sig_params (src, "signatureAlgorithm",
-					  issuer_key->pk_algorithm, dig,
-					  issuer_key->params,
-					  issuer_key->params_size);
+					  gnutls_privkey_get_pk_algorithm(issuer_key, NULL), dig);
   if (result < 0)
     {
       gnutls_assert ();
