@@ -53,15 +53,38 @@ static int _gnutls_verify_crl2 (gnutls_x509_crl_t crl,
 				int tcas_size, unsigned int flags,
 				unsigned int *output);
 
-/* Checks if two certs are identical.  Return 0 onn match. */
+/* Checks if two certs are identical.  Return 0 on match. */
 static int
 check_if_same_cert (gnutls_x509_crt_t cert1, gnutls_x509_crt_t cert2)
 {
   gnutls_datum_t cert1bin = { NULL, 0 }, cert2bin =
-  {
-  NULL, 0};
+  {NULL, 0};
   int result;
+  opaque serial1[128], serial2[128];
+  size_t serial1_size, serial2_size;
 
+  serial1_size = sizeof (serial1);
+  result = gnutls_x509_crt_get_serial (cert1, serial1, &serial1_size);
+  if (result < 0)
+    {
+      gnutls_assert ();
+      goto cmp;
+    }
+
+  serial2_size = sizeof (serial2);
+  result = gnutls_x509_crt_get_serial (cert2, serial2, &serial2_size);
+  if (result < 0)
+    {
+      gnutls_assert ();
+      goto cmp;
+    }
+
+  if (serial2_size != serial1_size || memcmp(serial1, serial2, serial1_size) != 0)
+    {
+      return 1;
+    }
+
+cmp:
   result = _gnutls_x509_der_encode (cert1->cert, "", &cert1bin, 0);
   if (result < 0)
     {
@@ -473,34 +496,36 @@ _gnutls_x509_verify_certificate (const gnutls_x509_crt_t * certificate_list,
    * self-signed E but already removed above), and we trust B, remove
    * B, C and D. */
   if (!(flags & GNUTLS_VERIFY_DO_NOT_ALLOW_SAME))
+    i = 0; /* also replace the first one */
+  else
+    i = 1; /* do not replace the first one */
+    
+  for (; i < clist_size; i++)
     {
-      for (i = 0; i < clist_size; i++)
-	{
-	  int j;
+      int j;
 
-	  for (j = 0; j < tcas_size; j++)
-	    {
-	      if (check_if_same_cert (certificate_list[i],
+      for (j = 0; j < tcas_size; j++)
+        {
+          if (check_if_same_cert (certificate_list[i],
 				      trusted_cas[j]) == 0)
-		{
-		  /* explicity time check for trusted CA that we remove from
-		   * list. GNUTLS_VERIFY_DISABLE_TRUSTED_TIME_CHECKS
-		   */
-		  if (!(flags & GNUTLS_VERIFY_DISABLE_TRUSTED_TIME_CHECKS)
+	    {
+	      /* explicity time check for trusted CA that we remove from
+	       * list. GNUTLS_VERIFY_DISABLE_TRUSTED_TIME_CHECKS
+	       */
+	      if (!(flags & GNUTLS_VERIFY_DISABLE_TRUSTED_TIME_CHECKS)
 		      && !(flags & GNUTLS_VERIFY_DISABLE_TIME_CHECKS))
-		    {
+                {
 		      status |= check_time (trusted_cas[j], now);
 		      if (status != 0)
 			{
 			  return status;
 			}
-		    }
-		  clist_size = i;
-		  break;
 		}
+              clist_size = i;
+	      break;
 	    }
-	  /* clist_size may have been changed which gets out of loop */
-	}
+        }
+        /* clist_size may have been changed which gets out of loop */
     }
 
   if (clist_size == 0)
@@ -800,8 +825,8 @@ dsa_verify_sig (const gnutls_datum_t * text,
 /* Verifies the signature data, and returns 0 if not verified,
  * or 1 otherwise.
  */
-static int
-verify_sig (const gnutls_datum_t * tbs,
+int
+pubkey_verify_sig (const gnutls_datum_t * tbs,
 	    const gnutls_datum_t * hash,
 	    const gnutls_datum_t * signature,
 	    gnutls_pk_algorithm_t pk, bigint_t * issuer_params,
@@ -839,57 +864,49 @@ verify_sig (const gnutls_datum_t * tbs,
     }
 }
 
-gnutls_digest_algorithm_t
-_gnutls_dsa_q_to_hash (bigint_t q)
+gnutls_digest_algorithm_t _gnutls_dsa_q_to_hash(bigint_t q)
 {
-  int bits = _gnutls_mpi_get_nbits (q);
+  int bits = _gnutls_mpi_get_nbits(q);
 
-  if (bits <= 160)
-    {
-      return GNUTLS_DIG_SHA1;
-    }
-  else if (bits <= 224)
-    {
-      return GNUTLS_DIG_SHA224;
-    }
-  else
-    {
-      return GNUTLS_DIG_SHA256;
-    }
+  if (bits <= 160) {
+    return GNUTLS_DIG_SHA1;
+  } else if (bits <= 224) {
+    return GNUTLS_DIG_SHA224;
+  } else {
+    return GNUTLS_DIG_SHA256;
+  }
 }
 
+/* This will return the appropriate hash to verify the given signature.
+ * If signature is NULL it will return an (or the) appropriate hash for
+ * the given parameters.
+ */
 int
 _gnutls_x509_verify_algorithm (gnutls_mac_algorithm_t * hash,
 			       const gnutls_datum_t * signature,
-			       const gnutls_x509_crt_t issuer)
+			       gnutls_pk_algorithm pk,
+			       bigint_t* issuer_params, unsigned int issuer_params_size)
 {
-  bigint_t issuer_params[MAX_PUBLIC_PARAMS_SIZE];
   opaque digest[MAX_HASH_SIZE];
   gnutls_datum_t decrypted;
-  int issuer_params_size;
   int digest_size;
-  int ret, i;
+  int ret;
 
-  issuer_params_size = MAX_PUBLIC_PARAMS_SIZE;
-  ret =
-    _gnutls_x509_crt_get_mpis (issuer, issuer_params, &issuer_params_size);
-  if (ret < 0)
-    {
-      gnutls_assert ();
-      return ret;
-    }
-
-  switch (gnutls_x509_crt_get_pk_algorithm (issuer, NULL))
+  switch(pk)
     {
     case GNUTLS_PK_DSA:
-
+      
       if (hash)
-	*hash = _gnutls_dsa_q_to_hash (issuer_params[1]);
+	*hash = _gnutls_dsa_q_to_hash(issuer_params[1]);
 
       ret = 0;
       break;
-
     case GNUTLS_PK_RSA:
+      if (signature == NULL) {/* return a sensible algorithm */
+        if (hash)
+          *hash = GNUTLS_DIG_SHA256;
+        return 0;
+      }
 
       ret =
 	_gnutls_pkcs1_rsa_decrypt (&decrypted, signature,
@@ -929,11 +946,6 @@ _gnutls_x509_verify_algorithm (gnutls_mac_algorithm_t * hash,
     }
 
 cleanup:
-  /* release allocated mpis */
-  for (i = 0; i < issuer_params_size; i++)
-    {
-      _gnutls_mpi_release (&issuer_params[i]);
-    }
 
   return ret;
 
@@ -966,7 +978,7 @@ _gnutls_x509_verify_signature (const gnutls_datum_t * tbs,
     }
 
   ret =
-    verify_sig (tbs, hash, signature,
+    pubkey_verify_sig (tbs, hash, signature,
 		gnutls_x509_crt_get_pk_algorithm (issuer, NULL),
 		issuer_params, issuer_params_size);
   if (ret < 0)
@@ -997,7 +1009,7 @@ _gnutls_x509_privkey_verify_signature (const gnutls_datum_t * tbs,
 {
   int ret;
 
-  ret = verify_sig (tbs, NULL, signature, issuer->pk_algorithm,
+  ret = pubkey_verify_sig (tbs, NULL, signature, issuer->pk_algorithm,
 		    issuer->params, issuer->params_size);
   if (ret < 0)
     {
