@@ -27,6 +27,7 @@
 # define X509_H
 
 #include <gnutls/x509.h>
+#include <gnutls/abstract.h>
 
 #include <libtasn1.h>
 
@@ -71,12 +72,12 @@ typedef struct gnutls_pkcs7_int
   ASN1_TYPE pkcs7;
 } gnutls_pkcs7_int;
 
-#define MAX_PRIV_PARAMS_SIZE GNUTLS_MAX_PK_PARAMS	/* ok for RSA and DSA */
+#define MAX_PRIV_PARAMS_SIZE GNUTLS_MAX_PK_PARAMS /* ok for RSA and DSA */
 
 /* parameters should not be larger than this limit */
 #define DSA_PRIVATE_PARAMS 5
 #define DSA_PUBLIC_PARAMS 4
-#define RSA_PRIVATE_PARAMS 6
+#define RSA_PRIVATE_PARAMS 8
 #define RSA_PUBLIC_PARAMS 2
 
 #if MAX_PRIV_PARAMS_SIZE - RSA_PRIVATE_PARAMS < 0
@@ -102,7 +103,12 @@ typedef struct gnutls_x509_privkey_int
    *      [4] is prime2 (q)
    *      [5] is coefficient (u == inverse of p mod q)
    *          note that other packages used inverse of q mod p,
-   *          so we need to perform conversions.
+   *          so we need to perform conversions on import/export
+   *          using fixup.
+   *      The following two are also not always available thus fixup
+   *      will generate them.
+   *	  [6] e1 == d mod (p-1)
+   *	  [7] e2 == d mod (q-1)
    * DSA: [0] is p
    *      [1] is q
    *      [2] is g
@@ -134,18 +140,12 @@ int _gnutls_x509_crl_get_raw_issuer_dn (gnutls_x509_crl_t crl,
 					gnutls_datum_t * dn);
 
 /* sign.c */
-int _gnutls_x509_sign (const gnutls_datum_t * tbs,
-		       gnutls_digest_algorithm_t hash,
-		       gnutls_x509_privkey_t signer,
-		       gnutls_datum_t * signature);
-int _gnutls_x509_sign_tbs (ASN1_TYPE cert, const char *tbs_name,
-			   gnutls_digest_algorithm_t hash,
-			   gnutls_x509_privkey_t signer,
-			   gnutls_datum_t * signature);
+int _gnutls_x509_get_tbs (ASN1_TYPE cert, const char *tbs_name,
+			   gnutls_datum_t * tbs);
 int _gnutls_x509_pkix_sign (ASN1_TYPE src, const char *src_name,
 			    gnutls_digest_algorithm_t,
 			    gnutls_x509_crt_t issuer,
-			    gnutls_x509_privkey_t issuer_key);
+			    gnutls_privkey_t issuer_key);
 
 /* dn.c */
 #define OID_X520_COUNTRY_NAME		"2.5.4.6"
@@ -185,9 +185,13 @@ int _gnutls_parse_general_name (ASN1_TYPE src, const char *src_name,
 /* verify.c */
 int gnutls_x509_crt_is_issuer (gnutls_x509_crt_t cert,
 			       gnutls_x509_crt_t issuer);
-int _gnutls_x509_verify_algorithm (gnutls_mac_algorithm_t * hash,
-				   const gnutls_datum_t * signature,
-				   const gnutls_x509_crt_t crt);
+
+int
+_gnutls_x509_verify_algorithm (gnutls_mac_algorithm_t * hash,
+			       const gnutls_datum_t * signature,
+			       gnutls_pk_algorithm pk,
+			       bigint_t* issuer_params, unsigned int issuer_params_size);
+
 int _gnutls_x509_verify_signature (const gnutls_datum_t * tbs,
 				   const gnutls_datum_t * hash,
 				   const gnutls_datum_t * signature,
@@ -307,8 +311,8 @@ int _gnutls_x509_write_uint32 (ASN1_TYPE node, const char *value,
 
 int _gnutls_x509_write_sig_params (ASN1_TYPE dst, const char *dst_name,
 				   gnutls_pk_algorithm_t pk_algorithm,
-				   gnutls_digest_algorithm_t,
-				   bigint_t * params, int params_size);
+				   gnutls_digest_algorithm_t);
+
 /* pkcs12.h */
 #include <gnutls/pkcs12.h>
 
