@@ -8,48 +8,64 @@ typedef struct _gaainfo gaainfo;
 
 struct _gaainfo
 {
-#line 131 "certtool.gaa"
+#line 165 "certtool.gaa"
 	int debug;
-#line 127 "certtool.gaa"
+#line 160 "certtool.gaa"
+	int pkcs11_trusted;
+#line 157 "certtool.gaa"
+	char* pkcs11_label;
+#line 154 "certtool.gaa"
+	int pkcs11_login;
+#line 146 "certtool.gaa"
+	int pkcs11_type;
+#line 143 "certtool.gaa"
+	char* pkcs11_url;
+#line 140 "certtool.gaa"
+	char* pkcs11_provider;
+#line 137 "certtool.gaa"
 	char *pkcs_cipher;
-#line 124 "certtool.gaa"
+#line 134 "certtool.gaa"
 	char *template;
-#line 121 "certtool.gaa"
+#line 131 "certtool.gaa"
 	char *infile;
-#line 118 "certtool.gaa"
+#line 128 "certtool.gaa"
 	char *outfile;
-#line 115 "certtool.gaa"
+#line 125 "certtool.gaa"
 	int quick_random;
-#line 112 "certtool.gaa"
+#line 122 "certtool.gaa"
+	char* sec_param;
+#line 119 "certtool.gaa"
 	int bits;
-#line 108 "certtool.gaa"
+#line 115 "certtool.gaa"
 	int outcert_format;
-#line 104 "certtool.gaa"
+#line 111 "certtool.gaa"
 	int incert_format;
-#line 101 "certtool.gaa"
+#line 108 "certtool.gaa"
 	int export;
-#line 98 "certtool.gaa"
+#line 105 "certtool.gaa"
 	char *hash;
-#line 95 "certtool.gaa"
+#line 102 "certtool.gaa"
 	int dsa;
-#line 92 "certtool.gaa"
+#line 99 "certtool.gaa"
 	int pkcs8;
-#line 85 "certtool.gaa"
+#line 92 "certtool.gaa"
 	int v1_cert;
-#line 82 "certtool.gaa"
+#line 89 "certtool.gaa"
 	int fix_key;
-#line 67 "certtool.gaa"
+#line 72 "certtool.gaa"
 	int crq_extensions;
-#line 54 "certtool.gaa"
+#line 57 "certtool.gaa"
 	char *pass;
-#line 51 "certtool.gaa"
+#line 54 "certtool.gaa"
 	char *ca;
-#line 48 "certtool.gaa"
+#line 51 "certtool.gaa"
 	char *ca_privkey;
-#line 45 "certtool.gaa"
+#line 48 "certtool.gaa"
 	char *cert;
-#line 42 "certtool.gaa"
+#line 45 "certtool.gaa"
 	char *request;
+#line 42 "certtool.gaa"
+	char *pubkey;
 #line 39 "certtool.gaa"
 	char *privkey;
 #line 17 "certtool.gaa"
