@@ -25,6 +25,8 @@
 #include <gnutls/x509.h>
 #include <gnutls/openpgp.h>
 #include <gnutls/pkcs12.h>
+#include <gnutls/pkcs11.h>
+#include <gnutls/abstract.h>
 
 #include <gcrypt.h>
 
@@ -50,7 +52,6 @@
 #include "certtool-common.h"
 
 static void print_crl_info (gnutls_x509_crl_t crl, FILE * out);
-int generate_prime (int bits, int how);
 void pkcs7_info (void);
 void crq_info (void);
 void smime_to_pkcs7 (void);
@@ -59,31 +60,25 @@ void generate_pkcs12 (void);
 void generate_pkcs8 (void);
 void verify_chain (void);
 void verify_crl (void);
+void pubkey_info (gnutls_x509_crt crt);
 void pgp_privkey_info (void);
 void pgp_ring_info (void);
-gnutls_x509_privkey_t load_private_key (int mand);
-gnutls_x509_crq_t load_request (void);
-gnutls_x509_privkey_t load_ca_private_key (void);
-gnutls_x509_crt_t load_ca_cert (void);
-gnutls_x509_crt_t load_cert (int mand);
-void certificate_info (void);
+void certificate_info (int);
 void pgp_certificate_info (void);
 void crl_info (void);
 void privkey_info (void);
-static void print_certificate_info (gnutls_x509_crt_t crt, FILE * out,
-				    unsigned int);
 static void gaa_parser (int argc, char **argv);
 void generate_self_signed (void);
 void generate_request (void);
 gnutls_x509_crt_t *load_cert_list (int mand, size_t * size);
+static void print_certificate_info (gnutls_x509_crt_t crt, FILE * out, unsigned int all);
 
 static void print_hex_datum (gnutls_datum_t * dat);
 
-
 static gaainfo info;
 FILE *outfile;
 FILE *infile;
-gnutls_digest_algorithm_t dig = GNUTLS_DIG_SHA1;
+gnutls_digest_algorithm_t default_dig;
 
 /* non interactive operation if set
  */
@@ -133,8 +128,11 @@ static void
 print_dsa_pkey (gnutls_datum_t * x, gnutls_datum_t * y, gnutls_datum_t * p,
 		gnutls_datum_t * q, gnutls_datum_t * g)
 {
-  fprintf (outfile, "private key:");
-  print_hex_datum (x);
+  if (x) 
+    {
+      fprintf (outfile, "private key:");
+      print_hex_datum (x);
+    }
   fprintf (outfile, "public key:");
   print_hex_datum (y);
   fprintf (outfile, "p:");
@@ -147,36 +145,96 @@ print_dsa_pkey (gnutls_datum_t * x, gnutls_datum_t * y, gnutls_datum_t * p,
 
 static void
 print_rsa_pkey (gnutls_datum_t * m, gnutls_datum_t * e, gnutls_datum_t * d,
-		gnutls_datum_t * p, gnutls_datum_t * q, gnutls_datum_t * u)
+		gnutls_datum_t * p, gnutls_datum_t * q, gnutls_datum_t * u,
+		gnutls_datum_t * exp1, gnutls_datum_t *exp2)
 {
   fprintf (outfile, "modulus:");
   print_hex_datum (m);
   fprintf (outfile, "public exponent:");
   print_hex_datum (e);
-  fprintf (outfile, "private exponent:");
-  print_hex_datum (d);
-  fprintf (outfile, "prime1:");
-  print_hex_datum (p);
-  fprintf (outfile, "prime2:");
-  print_hex_datum (q);
-  fprintf (outfile, "coefficient:");
-  print_hex_datum (u);
+  if (d) 
+    {
+      fprintf (outfile, "private exponent:");
+      print_hex_datum (d);
+      fprintf (outfile, "prime1:");
+      print_hex_datum (p);
+      fprintf (outfile, "prime2:");
+      print_hex_datum (q);
+      fprintf (outfile, "coefficient:");
+      print_hex_datum (u);
+      if (exp1 && exp2)
+        {
+          fprintf (outfile, "exp1:");
+          print_hex_datum (exp1);
+          fprintf (outfile, "exp2:");
+          print_hex_datum (exp2);
+        }
+    }
 }
 
+static gnutls_sec_param_t str_to_sec_param(const char* str)
+{
+  if (strcasecmp(str, "low")==0)
+    {
+      return GNUTLS_SEC_PARAM_LOW;
+    }
+  else if (strcasecmp(str, "normal")==0)
+    {
+      return GNUTLS_SEC_PARAM_NORMAL;
+    }
+  else if (strcasecmp(str, "high")==0)
+    {
+      return GNUTLS_SEC_PARAM_HIGH;
+    }
+  else if (strcasecmp(str, "ultra")==0)
+    {
+      return GNUTLS_SEC_PARAM_ULTRA;
+    }
+  else
+    {
+      fprintf(stderr, "Unknown security parameter string: %s\n", str);
+      exit(1);
+    }
+
+}
+
+int get_bits(gnutls_pk_algorithm_t key_type)
+{
+int bits;
+
+  if (info.bits != 0)
+    {
+      static int warned = 0;
+      
+      if (warned == 0) 
+        {
+          warned = 1;
+          fprintf(stderr, "** Note: Please use the --sec-param instead of --bits\n");
+        }
+      bits = info.bits;
+    }
+  else
+    {
+      if (info.sec_param)
+        {
+          bits = gnutls_sec_param_to_pk_bits(key_type, str_to_sec_param(info.sec_param));
+        }
+      else bits = gnutls_sec_param_to_pk_bits(key_type, GNUTLS_SEC_PARAM_NORMAL);
+    }
+
+  return bits;
+}
+
+
 static gnutls_x509_privkey_t
 generate_private_key_int (void)
 {
   gnutls_x509_privkey_t key;
-  int ret, key_type;
+  int ret, key_type, bits;
 
   if (info.dsa)
     {
       key_type = GNUTLS_PK_DSA;
-      if (info.bits > 1024)
-	{
-	  fprintf (stderr,
-		   "It is suggested for compatibility with other implementations to use DSA keys bit length 1024\n");
-	}
     }
   else
     key_type = GNUTLS_PK_RSA;
@@ -185,7 +243,9 @@ generate_private_key_int (void)
   if (ret < 0)
     error (EXIT_FAILURE, 0, "privkey_init: %s", gnutls_strerror (ret));
 
-  fprintf (stderr, "Generating a %d bit %s private key...\n", info.bits,
+  bits = get_bits(key_type);
+
+  fprintf (stderr, "Generating a %d bit %s private key...\n", get_bits(key_type),
 	   gnutls_pk_algorithm_get_name (key_type));
 
   if (info.quick_random == 0)
@@ -193,7 +253,7 @@ generate_private_key_int (void)
 	     "This might take several minutes depending on availability of randomness"
 	     " in /dev/random.\n");
 
-  ret = gnutls_x509_privkey_generate (key, key_type, info.bits, 0);
+  ret = gnutls_x509_privkey_generate (key, key_type, get_bits(key_type), 0);
   if (ret < 0)
     error (EXIT_FAILURE, 0, "privkey_generate: %s", gnutls_strerror (ret));
 
@@ -543,18 +603,6 @@ generate_certificate (gnutls_x509_privkey_t * ret_key,
 		error (EXIT_FAILURE, 0, "key_kp: %s",
 		       gnutls_strerror (result));
 	    }
-
-	  result = get_ipsec_ike_status ();
-	  if (result)
-	    {
-	      result =
-		gnutls_x509_crt_set_key_purpose_oid (crt,
-						     GNUTLS_KP_IPSEC_IKE,
-						     0);
-	      if (result < 0)
-		error (EXIT_FAILURE, 0, "key_kp: %s",
-		       gnutls_strerror (result));
-	    }
 	}
 
       if (usage != 0)
@@ -692,8 +740,27 @@ generate_crl (gnutls_x509_crt_t ca_crt)
   return crl;
 }
 
-void
-generate_self_signed (void)
+static gnutls_digest_algorithm_t get_dig(gnutls_x509_crt crt)
+{
+gnutls_digest_algorithm_t dig;
+int result;
+unsigned int mand;
+
+  result = gnutls_x509_crt_get_preferred_hash_algorithm(crt, &dig, &mand);
+  if (result < 0)
+    {
+	    error (EXIT_FAILURE, 0, "crl_preferred_hash_algorithm: %s",
+	       gnutls_strerror (result));
+    }
+
+  /* if algorithm allows alternatives */
+  if (mand == 0 && default_dig != GNUTLS_DIG_UNKNOWN)
+    dig = default_dig;
+
+  return dig;
+}
+
+void generate_self_signed (void)
 {
   gnutls_x509_crt_t crt;
   gnutls_x509_privkey_t key;
@@ -723,7 +790,7 @@ generate_self_signed (void)
 
   fprintf (stderr, "\n\nSigning certificate...\n");
 
-  result = gnutls_x509_crt_sign2 (crt, crt, key, dig, 0);
+  result = gnutls_x509_crt_sign2 (crt, crt, key, get_dig(crt), 0);
   if (result < 0)
     error (EXIT_FAILURE, 0, "crt_sign: %s", gnutls_strerror (result));
 
@@ -765,7 +832,7 @@ generate_signed_certificate (void)
 
   fprintf (stderr, "\n\nSigning certificate...\n");
 
-  result = gnutls_x509_crt_sign2 (crt, ca_crt, ca_key, dig, 0);
+  result = gnutls_x509_crt_sign2 (crt, ca_crt, ca_key, get_dig(ca_crt), 0);
   if (result < 0)
     error (EXIT_FAILURE, 0, "crt_sign: %s", gnutls_strerror (result));
 
@@ -799,7 +866,7 @@ generate_proxy_certificate (void)
 
   fprintf (stderr, "\n\nSigning certificate...\n");
 
-  result = gnutls_x509_crt_sign2 (crt, eecrt, eekey, dig, 0);
+  result = gnutls_x509_crt_sign2 (crt, eecrt, eekey, get_dig(eecrt), 0);
   if (result < 0)
     error (EXIT_FAILURE, 0, "crt_sign: %s", gnutls_strerror (result));
 
@@ -868,7 +935,7 @@ update_signed_certificate (void)
 
   fprintf (stderr, "\n\nSigning certificate...\n");
 
-  result = gnutls_x509_crt_sign2 (crt, ca_crt, ca_key, dig, 0);
+  result = gnutls_x509_crt_sign2 (crt, ca_crt, ca_key, get_dig(ca_crt), 0);
   if (result < 0)
     error (EXIT_FAILURE, 0, "crt_sign: %s", gnutls_strerror (result));
 
@@ -943,24 +1010,27 @@ gaa_parser (int argc, char **argv)
   else
     info.outcert_format = GNUTLS_X509_FMT_PEM;
 
+  default_dig = GNUTLS_DIG_UNKNOWN;
   if (info.hash != NULL)
     {
       if (strcasecmp (info.hash, "md5") == 0)
 	{
 	  fprintf (stderr,
 		   "Warning: MD5 is broken, and should not be used any more for digital signatures.\n");
-	  dig = GNUTLS_DIG_MD5;
+	  default_dig = GNUTLS_DIG_MD5;
 	}
       else if (strcasecmp (info.hash, "sha1") == 0)
-	dig = GNUTLS_DIG_SHA1;
+	default_dig = GNUTLS_DIG_SHA1;
       else if (strcasecmp (info.hash, "sha256") == 0)
-	dig = GNUTLS_DIG_SHA256;
+	default_dig = GNUTLS_DIG_SHA256;
+      else if (strcasecmp (info.hash, "sha224") == 0)
+	default_dig = GNUTLS_DIG_SHA224;
       else if (strcasecmp (info.hash, "sha384") == 0)
-	dig = GNUTLS_DIG_SHA384;
+	default_dig = GNUTLS_DIG_SHA384;
       else if (strcasecmp (info.hash, "sha512") == 0)
-	dig = GNUTLS_DIG_SHA512;
+	default_dig = GNUTLS_DIG_SHA512;
       else if (strcasecmp (info.hash, "rmd160") == 0)
-	dig = GNUTLS_DIG_RMD160;
+	default_dig = GNUTLS_DIG_RMD160;
       else
 	error (EXIT_FAILURE, 0, "invalid hash: %s", info.hash);
     }
@@ -972,19 +1042,6 @@ gaa_parser (int argc, char **argv)
       template_parse (info.template);
     }
 
-#ifdef gcry_fips_mode_active
-  /* Libgcrypt manual says that gcry_version_check must be called
-     before calling gcry_fips_mode_active. */
-  gcry_check_version (NULL);
-  if (gcry_fips_mode_active ())
-    {
-      ret = gnutls_register_md5_handler ();
-      if (ret)
-	fprintf (stderr, "gnutls_register_md5_handler: %s\n",
-		 gnutls_strerror (ret));
-    }
-#endif
-
   gnutls_global_set_log_function (tls_log_func);
   gnutls_global_set_log_level (info.debug);
   if (info.debug > 1)
@@ -992,13 +1049,29 @@ gaa_parser (int argc, char **argv)
 
   if ((ret = gnutls_global_init ()) < 0)
     error (EXIT_FAILURE, 0, "global_init: %s", gnutls_strerror (ret));
+    
+  if (info.pkcs11_provider != NULL)
+    {
+      ret = gnutls_pkcs11_init(GNUTLS_PKCS11_FLAG_MANUAL, NULL);
+      if (ret < 0)
+        fprintf (stderr, "pkcs11_init: %s", gnutls_strerror (ret));
+      else {
+        ret = gnutls_pkcs11_add_provider(info.pkcs11_provider, NULL);
+        if (ret < 0)
+          error (EXIT_FAILURE, 0, "pkcs11_add_provider: %s", gnutls_strerror (ret));
+      }
+    }
+  else
+    {
+      ret = gnutls_pkcs11_init(GNUTLS_PKCS11_FLAG_AUTO, NULL);
+      if (ret < 0)
+        fprintf (stderr, "pkcs11_init: %s", gnutls_strerror (ret));
+    }
+      
 
   if ((ret = gnutls_global_init_extra ()) < 0)
     error (EXIT_FAILURE, 0, "global_init_extra: %s", gnutls_strerror (ret));
 
-  if (info.quick_random != 0)
-    gcry_control (GCRYCTL_ENABLE_QUICK_RANDOM, 0);
-
   switch (info.action)
     {
     case ACTION_SELF_SIGNED:
@@ -1008,7 +1081,10 @@ gaa_parser (int argc, char **argv)
       generate_private_key ();
       break;
     case ACTION_CERT_INFO:
-      certificate_info ();
+      certificate_info (0);
+      break;
+    case ACTION_CERT_PUBKEY:
+      certificate_info (1);
       break;
     case ACTION_GENERATE_REQUEST:
       generate_request ();
@@ -1022,6 +1098,9 @@ gaa_parser (int argc, char **argv)
     case ACTION_PRIVKEY_INFO:
       privkey_info ();
       break;
+    case ACTION_PUBKEY_INFO:
+      pubkey_info (NULL);
+      break;
     case ACTION_UPDATE_CERTIFICATE:
       update_signed_certificate ();
       break;
@@ -1032,10 +1111,10 @@ gaa_parser (int argc, char **argv)
       pkcs12_info ();
       break;
     case ACTION_GENERATE_DH:
-      generate_prime (info.bits, 1);
+      generate_prime (1);
       break;
     case ACTION_GET_DH:
-      generate_prime (info.bits, 0);
+      generate_prime (0);
       break;
     case ACTION_CRL_INFO:
       crl_info ();
@@ -1058,6 +1137,21 @@ gaa_parser (int argc, char **argv)
     case ACTION_GENERATE_PKCS8:
       generate_pkcs8 ();
       break;
+    case ACTION_PKCS11_LIST:
+      pkcs11_list(outfile, info.pkcs11_url, info.pkcs11_type, info.pkcs11_login);
+      break;
+    case ACTION_PKCS11_TOKENS:
+      pkcs11_token_list(outfile);
+      break;
+    case ACTION_PKCS11_EXPORT_URL:
+      pkcs11_export(outfile, info.pkcs11_url, info.pkcs11_login);
+      break;
+    case ACTION_PKCS11_WRITE_URL:
+      pkcs11_write(outfile, info.pkcs11_url, info.pkcs11_label, info.pkcs11_trusted, info.pkcs11_login);
+      break;
+    case ACTION_PKCS11_DELETE_URL:
+      pkcs11_delete(outfile, info.pkcs11_url, batch, info.pkcs11_login);
+      break;
 #ifdef ENABLE_OPENPGP
     case ACTION_PGP_INFO:
       pgp_certificate_info ();
@@ -1077,11 +1171,14 @@ gaa_parser (int argc, char **argv)
       exit (0);
     }
   fclose (outfile);
+  
+  gnutls_pkcs11_deinit();
+  gnutls_global_deinit();
 }
 
 #define MAX_CRTS 500
 void
-certificate_info (void)
+certificate_info (int pubkey)
 {
   gnutls_x509_crt_t crt[MAX_CRTS];
   size_t size;
@@ -1131,8 +1228,11 @@ certificate_info (void)
 				    &size);
       if (ret < 0)
 	error (EXIT_FAILURE, 0, "export error: %s", gnutls_strerror (ret));
+
       fwrite (buffer, 1, size, outfile);
 
+      if (pubkey) pubkey_info(crt[i]);
+
       gnutls_x509_crt_deinit (crt[i]);
     }
 }
@@ -1271,7 +1371,7 @@ pgp_privkey_info (void)
 	    fprintf (stderr, "Error in key RSA data export: %s\n",
 		     gnutls_strerror (ret));
 	  else
-	    print_rsa_pkey (&m, &e, &d, &p, &q, &u);
+	    print_rsa_pkey (&m, &e, &d, &p, &q, &u, NULL, NULL);
 
 	}
       else if (ret == GNUTLS_PK_DSA)
@@ -1552,12 +1652,20 @@ privkey_info (void)
   /* If we failed to import the certificate previously try PKCS #8 */
   if (info.pkcs8 || ret == GNUTLS_E_BASE64_UNEXPECTED_HEADER_ERROR)
     {
-      if (info.pass)
-	pass = info.pass;
-      else
-	pass = get_pass ();
+      /* first try to import the key without asking any password */
       ret = gnutls_x509_privkey_import_pkcs8 (key, &pem,
-					      info.incert_format, pass, 0);
+					      info.incert_format,
+					      NULL, GNUTLS_PKCS_PLAIN);
+      if (ret < 0)
+	{
+	  if (info.pass)
+	    pass = info.pass;
+	  else
+	    pass = get_pass ();
+	  ret = gnutls_x509_privkey_import_pkcs8 (key, &pem,
+						  info.incert_format,
+						  pass, 0);
+	}
     }
   if (ret < 0)
     error (EXIT_FAILURE, 0, "import error: %s", gnutls_strerror (ret));
@@ -1575,21 +1683,23 @@ privkey_info (void)
    */
   if (ret == GNUTLS_PK_RSA)
     {
-      gnutls_datum_t m, e, d, p, q, u;
+      gnutls_datum_t m, e, d, p, q, u, exp1, exp2;
 
-      ret = gnutls_x509_privkey_export_rsa_raw (key, &m, &e, &d, &p, &q, &u);
+      ret = gnutls_x509_privkey_export_rsa_raw2 (key, &m, &e, &d, &p, &q, &u, &exp1, &exp2);
       if (ret < 0)
 	fprintf (stderr, "Error in key RSA data export: %s\n",
 		 gnutls_strerror (ret));
       else
 	{
-	  print_rsa_pkey (&m, &e, &d, &p, &q, &u);
+	  print_rsa_pkey (&m, &e, &d, &p, &q, &u, &exp1, &exp2);
 	  gnutls_free (m.data);
 	  gnutls_free (e.data);
 	  gnutls_free (d.data);
 	  gnutls_free (p.data);
 	  gnutls_free (q.data);
 	  gnutls_free (u.data);
+	  gnutls_free (exp1.data);
+	  gnutls_free (exp2.data);
 	}
     }
   else if (ret == GNUTLS_PK_DSA)
@@ -1641,6 +1751,51 @@ privkey_info (void)
   gnutls_x509_privkey_deinit (key);
 }
 
+/* Load a public key.
+ * @mand should be non zero if it is required to read a public key.
+ */
+gnutls_pubkey_t
+load_pubkey (int mand)
+{
+  gnutls_pubkey_t key;
+  int ret;
+  gnutls_datum_t dat;
+  size_t size;
+
+  if (!info.pubkey && !mand)
+    return NULL;
+
+  if (info.pubkey == NULL)
+    error (EXIT_FAILURE, 0, "missing --load-pubkey");
+
+  ret = gnutls_pubkey_init (&key);
+  if (ret < 0)
+    error (EXIT_FAILURE, 0, "privkey_init: %s", gnutls_strerror (ret));
+
+  dat.data = read_binary_file (info.pubkey, &size);
+  dat.size = size;
+
+  if (!dat.data)
+    error (EXIT_FAILURE, errno, "reading --load-pubkey: %s", info.pubkey);
+
+  ret = gnutls_pubkey_import (key, &dat, info.incert_format);
+
+  free (dat.data);
+
+  if (ret == GNUTLS_E_BASE64_UNEXPECTED_HEADER_ERROR)
+    {
+      error (EXIT_FAILURE, 0,
+	     "import error: could not find a valid PEM header; "
+	     "check if your key has the PUBLIC KEY header");
+    }
+
+  if (ret < 0)
+    error (EXIT_FAILURE, 0, "importing --load-pubkey: %s: %s",
+	   info.pubkey, gnutls_strerror (ret));
+
+  return key;
+}
+
 
 /* Load the private key.
  * @mand should be non zero if it is required to read a private key.
@@ -2004,15 +2159,6 @@ generate_request (void)
 	      if (ret < 0)
 		error (EXIT_FAILURE, 0, "key_kp: %s", gnutls_strerror (ret));
 	    }
-
-	  ret = get_ipsec_ike_status ();
-	  if (ret)
-	    {
-	      ret = gnutls_x509_crq_set_key_purpose_oid
-		(crq, GNUTLS_KP_IPSEC_IKE, 0);
-	      if (ret < 0)
-		error (EXIT_FAILURE, 0, "key_kp: %s", gnutls_strerror (ret));
-	    }
 	}
 
       ret = gnutls_x509_crq_set_key_usage (crq, usage);
@@ -2991,3 +3137,142 @@ certtool_version (void)
   version_etc (stdout, program_name, p, gnutls_check_version (NULL),
 	       "Nikos Mavrogiannopoulos", "Simon Josefsson", (char *) NULL);
 }
+
+static void print_key_usage(FILE* outfile, unsigned int usage)
+{
+  if (usage & GNUTLS_KEY_DIGITAL_SIGNATURE) {
+    fprintf(outfile, "\tDigital signature.\n");
+  }
+
+  if (usage & GNUTLS_KEY_NON_REPUDIATION) {
+    fprintf(outfile, "\tNon repudiation.\n");
+  }
+
+  if (usage & GNUTLS_KEY_KEY_ENCIPHERMENT) {
+    fprintf(outfile, "\tKey encipherment.\n");
+  }
+
+  if (usage & GNUTLS_KEY_DATA_ENCIPHERMENT) {
+    fprintf(outfile, "\tData encipherment.\n");
+  }
+
+  if (usage & GNUTLS_KEY_KEY_AGREEMENT) {
+    fprintf(outfile, "\tKey agreement.\n");
+  }
+
+  if (usage & GNUTLS_KEY_KEY_CERT_SIGN) {
+    fprintf(outfile, "\tCertificate signing.\n");
+  }
+
+  if (usage & GNUTLS_KEY_NON_REPUDIATION) {
+    fprintf(outfile, "\tCRL signing.\n");  
+  }
+
+  if (usage & GNUTLS_KEY_ENCIPHER_ONLY) {
+    fprintf(outfile, "\tKey encipher only.\n");
+  }
+
+  if (usage & GNUTLS_KEY_DECIPHER_ONLY) {
+    fprintf(outfile, "\tKey decipher only.\n");
+  }
+}
+
+void pubkey_info (gnutls_x509_crt crt)
+{
+  gnutls_pubkey_t pubkey;
+  unsigned int bits, usage;
+  int ret;
+  size_t size;
+  const char* cprint;
+
+  ret = gnutls_pubkey_init(&pubkey);
+  if (ret < 0) {
+    error (EXIT_FAILURE, 0, "pubkey_init: %s", gnutls_strerror (ret));
+  }
+
+  if (crt == NULL) {
+    crt = load_cert(0);
+  }
+  
+  if (crt != NULL) {
+    ret = gnutls_pubkey_import_x509(pubkey, crt, 0);
+    if (ret < 0) {
+      error (EXIT_FAILURE, 0, "pubkey_import_x509: %s", gnutls_strerror (ret));
+    }
+  } else {
+    pubkey = load_pubkey(1);
+  }
+
+  fprintf (outfile, "Public Key Info:\n\n");
+  ret = gnutls_pubkey_get_pk_algorithm (pubkey, &bits);
+  fprintf (outfile, "Public Key Algorithm: ");
+
+  cprint = gnutls_pk_algorithm_get_name (ret);
+  fprintf (outfile, "%s (%u bits)\n", cprint ? cprint : "Unknown", bits);
+
+
+  /* Print the raw public and private keys
+   */
+  if (ret == GNUTLS_PK_RSA)
+    {
+      gnutls_datum_t m, e;
+
+      ret = gnutls_pubkey_get_pk_rsa_raw (pubkey, &m, &e);
+      if (ret < 0)
+	fprintf (stderr, "Error in key RSA data export: %s\n",
+		 gnutls_strerror (ret));
+      else
+	{
+	  print_rsa_pkey (&m, &e, NULL, NULL, NULL, NULL, NULL, NULL);
+	  gnutls_free (m.data);
+	  gnutls_free (e.data);
+	}
+    }
+  else if (ret == GNUTLS_PK_DSA)
+    {
+      gnutls_datum_t p, q, g, y;
+
+      ret = gnutls_pubkey_get_pk_dsa_raw (pubkey, &p, &q, &g, &y);
+      if (ret < 0)
+	fprintf (stderr, "Error in key DSA data export: %s\n",
+		 gnutls_strerror (ret));
+      else
+	{
+	  print_dsa_pkey (NULL, &y, &p, &q, &g);
+	  gnutls_free (y.data);
+	  gnutls_free (p.data);
+	  gnutls_free (q.data);
+	  gnutls_free (g.data);
+	}
+    }
+
+  ret = gnutls_pubkey_get_key_usage (pubkey, &usage);
+  if (ret < 0) {
+    error (EXIT_FAILURE, 0, "pubkey_get_key_usage: %s", gnutls_strerror (ret));
+  }
+  
+  fprintf (outfile, "Public Key Usage:\n");
+  print_key_usage(outfile, usage);
+
+  fprintf (outfile, "\n");
+
+  size = sizeof (buffer);
+  if ((ret = gnutls_pubkey_get_key_id (pubkey, 0, buffer, &size)) < 0)
+    {
+      fprintf (stderr, "Error in key id calculation: %s\n",
+	       gnutls_strerror (ret));
+    }
+  else
+    {
+      fprintf (outfile, "Public Key ID: %s\n", raw_to_string (buffer, size));
+    }
+
+  size = sizeof (buffer);
+  ret = gnutls_pubkey_export (pubkey, GNUTLS_X509_FMT_PEM, buffer, &size);
+  if (ret < 0)
+    error (EXIT_FAILURE, 0, "export error: %s", gnutls_strerror (ret));
+
+  fprintf (outfile, "\n%s\n", buffer);
+
+  gnutls_pubkey_deinit (pubkey);
+}
